# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numba.core import cgutils, types
from numba.core.ir_utils import legalize_names

from numba_dppy import numpy_usm_shared as nus
from numba_dppy import utils
from numba_dppy.driver import DpctlCAPIFnBuilder
from numba_dppy.driver._helpers import numba_type_to_dpctl_typenum


class KernelLaunchOps:
    """
    KernelLaunchOps(lowerer, cres, num_inputs)
    Defines a set of functions to launch a SYCL kernel on the "current queue"
    as defined in the dpctl queue manager.
    """

    def _form_kernel_arg_and_arg_ty(self, val, ty):
        kernel_arg_dst = self.builder.gep(
            self.kernel_arg_array,
            [self.context.get_constant(types.int32, self.cur_arg)],
        )
        kernel_arg_ty_dst = self.builder.gep(
            self.kernel_arg_ty_array,
            [self.context.get_constant(types.int32, self.cur_arg)],
        )
        self.cur_arg += 1
        self.builder.store(val, kernel_arg_dst)
        self.builder.store(ty, kernel_arg_ty_dst)

    def __init__(self, lowerer, kernel, num_inputs):
        """Create an instance of KernelLaunchOps.

        Args:
            lowerer: The Numba Lowerer that will be used to generate the code.
            kernel: The SYCL kernel for which we are generating the code.
            num_inputs: The number of arguments to the kernels.
        """
        self.lowerer = lowerer
        self.context = self.lowerer.context
        self.builder = self.lowerer.builder
        self.kernel = kernel
        self.kernel_addr = self.kernel.addressof_ref()
        self.total_kernel_args = 0
        self.cur_arg = 0
        self.num_inputs = num_inputs
        # list of buffer that needs to comeback to host
        self.write_buffs = []
        # list of buffer that does not need to comeback to host
        self.read_only_buffs = []

    def get_current_queue(self):
        """Allocates memory on the stack to store the current queue from dpctl.

        A SYCL queue is needed to allocate USM memory and submit a kernel. This
        function gets the queue returned by ``DPCTLQueueMgr_GetCurrentQueue``
        function and stores it on the stack. The queue should be freed properly
        after returning from the kernel.

        Return: A LLVM Value storing the pointer to the SYCL queue returned
                by ``DPCTLQueueMgr_GetCurrentQueue``.

        """
        sycl_queue_val = cgutils.alloca_once(
            self.builder, utils.get_llvm_type(context=self.context, type=types.voidptr)
        )
        fn = DpctlCAPIFnBuilder.get_dpctl_queuemgr_get_current_queue(
            builder=self.builder, context=self.context
        )
        self.builder.store(self.builder.call(fn, []), sycl_queue_val)
        return sycl_queue_val

    def free_queue(self, sycl_queue_val):
        """
        Frees the ``DPCTLSyclQueueRef`` pointer that was used to launch the
        kernels.

        Args:
            sycl_queue_val: The SYCL queue pointer to be freed.
        """
        fn = DpctlCAPIFnBuilder.get_dpctl_queue_delete(
            builder=self.builder, context=self.context
        )
        self.builder.call(fn, [self.builder.load(sycl_queue_val)])

    def allocate_kernel_arg_array(self, num_kernel_args):
        """Allocates an array to store the LLVM Value for every kernel argument.

        Args:
            num_kernel_args : The number of kernel arguments for the kernel.

        """

        self.total_kernel_args = num_kernel_args

        # we need a kernel arg array to enqueue
        self.kernel_arg_array = cgutils.alloca_once(
            self.builder,
            utils.get_llvm_type(context=self.context, type=types.voidptr),
            size=self.context.get_constant(types.uintp, num_kernel_args),
            name="kernel_arg_array",
        )

        self.kernel_arg_ty_array = cgutils.alloca_once(
            self.builder,
            utils.LLVMTypes.int32_t,
            size=self.context.get_constant(types.uintp, num_kernel_args),
            name="kernel_arg_ty_array",
        )

    def process_kernel_arg(
        self, var, llvm_arg, arg_type, index, modified_arrays, sycl_queue_val
    ):
        """
        process_kernel_arg(var, llvm_arg, arg_type, index, modified_arrays,
        sycl_queue_val)
        Creates an LLVM Value for each kernel argument.

        Args:
            var : A kernel argument represented as a Numba type.
            llvm_arg : Only used for array arguments and points to the LLVM
                       value previously allocated to store the array arg.
            arg_type : The Numba type for the argument.
            index : The poisition of the argument in the list of arguments.
            modified_arrays : The list of array arguments that are written to
                              inside the kernel. The list is used to check if
                              the argument is read-only or not.

        Raises:
            NotImplementedError: If an unsupported type of kernel argument is
                                 encountered.
        """
        if isinstance(arg_type, types.npytypes.Array):
            if llvm_arg is None:
                raise NotImplementedError(arg_type, var)

            storage = cgutils.alloca_once(self.builder, utils.LLVMTypes.int64_t)
            self.builder.store(self.context.get_constant(types.int64, 0), storage)
            ty = numba_type_to_dpctl_typenum(context=self.context, type=types.int64)
            self._form_kernel_arg_and_arg_ty(
                self.builder.bitcast(
                    storage,
                    utils.get_llvm_type(context=self.context, type=types.voidptr),
                ),
                ty,
            )

            storage = cgutils.alloca_once(self.builder, utils.LLVMTypes.int64_t)
            self.builder.store(self.context.get_constant(types.int64, 0), storage)
            ty = numba_type_to_dpctl_typenum(context=self.context, type=types.int64)
            self._form_kernel_arg_and_arg_ty(
                self.builder.bitcast(
                    storage,
                    utils.get_llvm_type(context=self.context, type=types.voidptr),
                ),
                ty,
            )

            # Handle array size
            array_size_member = self.builder.gep(
                llvm_arg,
                [
                    self.context.get_constant(types.int32, 0),
                    self.context.get_constant(types.int32, 2),
                ],
            )

            ty = numba_type_to_dpctl_typenum(context=self.context, type=types.int64)
            self._form_kernel_arg_and_arg_ty(
                self.builder.bitcast(
                    array_size_member,
                    utils.get_llvm_type(context=self.context, type=types.voidptr),
                ),
                ty,
            )

            # Handle itemsize
            item_size_member = self.builder.gep(
                llvm_arg,
                [
                    self.context.get_constant(types.int32, 0),
                    self.context.get_constant(types.int32, 3),
                ],
            )

            ty = numba_type_to_dpctl_typenum(context=self.context, type=types.int64)
            self._form_kernel_arg_and_arg_ty(
                self.builder.bitcast(
                    item_size_member,
                    utils.get_llvm_type(context=self.context, type=types.voidptr),
                ),
                ty,
            )

            # Calculate total buffer size
            total_size = cgutils.alloca_once(
                self.builder,
                utils.get_llvm_type(context=self.context, type=types.intp),
                size=utils.get_one(context=self.context),
                name="total_size" + str(self.cur_arg),
            )
            self.builder.store(
                self.builder.sext(
                    self.builder.mul(
                        self.builder.load(array_size_member),
                        self.builder.load(item_size_member),
                    ),
                    utils.get_llvm_type(context=self.context, type=types.intp),
                ),
                total_size,
            )

            # Handle data
            data_member = self.builder.gep(
                llvm_arg,
                [
                    self.context.get_constant(types.int32, 0),
                    self.context.get_constant(types.int32, 4),
                ],
            )

            # names are replaced using legalize names, we have to do the same
            # here for them to match.
            legal_names = legalize_names([var])
            ty = numba_type_to_dpctl_typenum(context=self.context, type=types.voidptr)

            if isinstance(arg_type, nus.UsmSharedArrayType):
                self._form_kernel_arg_and_arg_ty(
                    self.builder.bitcast(
                        self.builder.load(data_member),
                        utils.get_llvm_type(context=self.context, type=types.voidptr),
                    ),
                    ty,
                )
            else:
                malloc_fn = DpctlCAPIFnBuilder.get_dpctl_malloc_shared(
                    builder=self.builder, context=self.context
                )
                memcpy_fn = DpctlCAPIFnBuilder.get_dpctl_queue_memcpy(
                    builder=self.builder, context=self.context
                )
                event_del_fn = DpctlCAPIFnBuilder.get_dpctl_event_delete(
                    builder=self.builder, context=self.context
                )
                event_wait_fn = DpctlCAPIFnBuilder.get_dpctl_event_wait(
                    builder=self.builder, context=self.context
                )

                # Not known to be USM so we need to copy to USM.
                buffer_name = "buffer_ptr" + str(self.cur_arg)
                # Create void * to hold new USM buffer.
                buffer_ptr = cgutils.alloca_once(
                    self.builder,
                    utils.get_llvm_type(context=self.context, type=types.voidptr),
                    name=buffer_name,
                )
                # Setup the args to the USM allocator, size and SYCL queue.
                args = [
                    self.builder.load(total_size),
                    self.builder.load(sycl_queue_val),
                ]
                # Call USM shared allocator and store in buffer_ptr.
                self.builder.store(self.builder.call(malloc_fn, args), buffer_ptr)

                if legal_names[var] in modified_arrays:
                    self.write_buffs.append((buffer_ptr, total_size, data_member))
                else:
                    self.read_only_buffs.append((buffer_ptr, total_size, data_member))

                # We really need to detect when an array needs to be copied over
                if index < self.num_inputs:
                    args = [
                        self.builder.load(sycl_queue_val),
                        self.builder.load(buffer_ptr),
                        self.builder.bitcast(
                            self.builder.load(data_member),
                            utils.get_llvm_type(
                                context=self.context, type=types.voidptr
                            ),
                        ),
                        self.builder.load(total_size),
                    ]
                    event_ref = self.builder.call(memcpy_fn, args)
                    self.builder.call(event_wait_fn, [event_ref])
                    self.builder.call(event_del_fn, [event_ref])

                self._form_kernel_arg_and_arg_ty(self.builder.load(buffer_ptr), ty)

            # Handle shape
            shape_member = self.builder.gep(
                llvm_arg,
                [
                    self.context.get_constant(types.int32, 0),
                    self.context.get_constant(types.int32, 5),
                ],
            )

            for this_dim in range(arg_type.ndim):
                shape_entry = self.builder.gep(
                    shape_member,
                    [
                        self.context.get_constant(types.int32, 0),
                        self.context.get_constant(types.int32, this_dim),
                    ],
                )
                ty = numba_type_to_dpctl_typenum(context=self.context, type=types.int64)
                self._form_kernel_arg_and_arg_ty(
                    self.builder.bitcast(
                        shape_entry,
                        utils.get_llvm_type(context=self.context, type=types.voidptr),
                    ),
                    ty,
                )

            # Handle strides
            stride_member = self.builder.gep(
                llvm_arg,
                [
                    self.context.get_constant(types.int32, 0),
                    self.context.get_constant(types.int32, 6),
                ],
            )

            for this_stride in range(arg_type.ndim):
                stride_entry = self.builder.gep(
                    stride_member,
                    [
                        self.context.get_constant(types.int32, 0),
                        self.context.get_constant(types.int32, this_stride),
                    ],
                )

                ty = numba_type_to_dpctl_typenum(context=self.context, type=types.int64)
                self._form_kernel_arg_and_arg_ty(
                    self.builder.bitcast(
                        stride_entry,
                        utils.get_llvm_type(context=self.context, type=types.voidptr),
                    ),
                    ty,
                )

        else:
            ty = numba_type_to_dpctl_typenum(context=self.context, type=arg_type)
            self._form_kernel_arg_and_arg_ty(
                self.builder.bitcast(
                    llvm_arg,
                    utils.get_llvm_type(context=self.context, type=types.voidptr),
                ),
                ty,
            )

    def enqueue_kernel_and_copy_back(self, dim_bounds, sycl_queue_val):
        """
        enqueue_kernel_and_copy_back(dim_bounds, sycl_queue_val)
        Submits the kernel to the specified queue, waits and then copies
        back any results to the host.

        Args:
            dim_bounds : An array of three tuple representing the starting
                         offset, end offset and the stride (step) for each
                         dimension of the input arrays. Every array in a parfor
                         is of the same dimensionality and shape, thus ensuring
                         the bounds are the same.
            sycl_queue_val : The SYCL queue on which the kernel is
                             submitted.
        """
        submit_fn = DpctlCAPIFnBuilder.get_dpctl_queue_submit_range(
            builder=self.builder, context=self.context
        )
        queue_wait_fn = DpctlCAPIFnBuilder.get_dpctl_queue_wait(
            builder=self.builder, context=self.context
        )
        event_del_fn = DpctlCAPIFnBuilder.get_dpctl_event_delete(
            builder=self.builder, context=self.context
        )
        memcpy_fn = DpctlCAPIFnBuilder.get_dpctl_queue_memcpy(
            builder=self.builder, context=self.context
        )
        free_fn = DpctlCAPIFnBuilder.get_dpctl_free_with_queue(
            builder=self.builder, context=self.context
        )
        event_wait_fn = DpctlCAPIFnBuilder.get_dpctl_event_wait(
            builder=self.builder, context=self.context
        )

        # the assumption is loop_ranges will always be less than or equal to 3
        # dimensions
        num_dim = len(dim_bounds) if len(dim_bounds) < 4 else 3

        # form the global range
        global_range = cgutils.alloca_once(
            self.builder,
            utils.get_llvm_type(context=self.context, type=types.uintp),
            size=self.context.get_constant(types.uintp, num_dim),
            name="global_range",
        )

        intp_t = utils.get_llvm_type(context=self.context, type=types.intp)
        intp_ptr_t = utils.get_llvm_ptr_type(intp_t)

        for i in range(num_dim):
            start, stop, step = dim_bounds[i]
            if stop.type != utils.LLVMTypes.int64_t:
                stop = self.builder.sext(stop, utils.LLVMTypes.int64_t)

            # we reverse the global range to account for how sycl and opencl
            # range differs
            self.builder.store(
                stop,
                self.builder.gep(
                    global_range,
                    [self.context.get_constant(types.uintp, (num_dim - 1) - i)],
                ),
            )

        args = [
            self.builder.inttoptr(
                self.context.get_constant(types.uintp, self.kernel_addr),
                utils.get_llvm_type(context=self.context, type=types.voidptr),
            ),
            self.builder.load(sycl_queue_val),
            self.kernel_arg_array,
            self.kernel_arg_ty_array,
            self.context.get_constant(types.uintp, self.total_kernel_args),
            self.builder.bitcast(global_range, intp_ptr_t),
            self.context.get_constant(types.uintp, num_dim),
            self.builder.bitcast(
                utils.create_null_ptr(builder=self.builder, context=self.context),
                utils.get_llvm_type(context=self.context, type=types.voidptr),
            ),
            self.context.get_constant(types.uintp, 0),
        ]

        # Submit the kernel
        event_ref = self.builder.call(submit_fn, args)

        # Add a wait on the queue
        self.builder.call(queue_wait_fn, [self.builder.load(sycl_queue_val)])

        # Note that the dpctl_queue_wait call waits on the event and then
        # decrements the ref count of the sycl::event C++ object. However, the
        # event object returned by the get_dpctl_queue_submit_range call still
        # needs to be explicitly deleted to free up the event object properly.
        self.builder.call(event_del_fn, [event_ref])

        # read buffers back to host
        for write_buff in self.write_buffs:
            buffer_ptr, total_size, data_member = write_buff
            args = [
                self.builder.load(sycl_queue_val),
                self.builder.bitcast(
                    self.builder.load(data_member),
                    utils.get_llvm_type(context=self.context, type=types.voidptr),
                ),
                self.builder.load(buffer_ptr),
                self.builder.load(total_size),
            ]
            # FIXME: In future, when the DctlQueue_Memcpy is made non-blocking
            # the returned event should be explicitly freed by calling
            # get_dpctl_event_delete.
            event_ref = self.builder.call(memcpy_fn, args)
            self.builder.call(event_wait_fn, [event_ref])
            self.builder.call(event_del_fn, [event_ref])

            self.builder.call(
                free_fn,
                [
                    self.builder.load(buffer_ptr),
                    self.builder.load(sycl_queue_val),
                ],
            )

        for read_buff in self.read_only_buffs:
            buffer_ptr, total_size, data_member = read_buff
            self.builder.call(
                free_fn,
                [
                    self.builder.load(buffer_ptr),
                    self.builder.load(sycl_queue_val),
                ],
            )



################################################################################
#                                 Numba-DPPY
#
# Copyright 2020-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import dpctl
import numpy as np
import pytest
from numba import njit

import numba_dppy as dppy
from numba_dppy.tests._helper import dpnp_debug, is_gen12

from ._helper import wrapper_function
from .dpnp_skip_test import dpnp_skip_test as skip_test

list_of_filter_strs = [
    "opencl:gpu:0",
    "level_zero:gpu:0",
    "opencl:cpu:0",
]


@pytest.fixture(params=list_of_filter_strs)
def filter_str(request):
    return request.param


list_of_int_dtypes = [
    np.int32,
    np.int64,
]


list_of_float_dtypes = [
    np.float32,
    np.float64,
]


@pytest.fixture(params=list_of_int_dtypes + list_of_float_dtypes)
def input_array(request):
    # The size of input and out arrays to be used
    N = 100
    a = np.array(np.random.random(N), request.param)
    return a


@pytest.fixture(params=list_of_float_dtypes)
def input_nan_array(request):
    # The size of input and out arrays to be used
    N = 100
    a = np.array(np.random.random(N), request.param)
    for i in range(5):
        a[N - 1 - i] = np.nan
    return a


list_of_shape = [
    (100),
    (50, 2),
    (10, 5, 2),
]


@pytest.fixture(params=list_of_shape)
def get_shape(request):
    return request.param


list_of_unary_ops = [
    "sum",
    "prod",
    "cumsum",
    "cumprod",
]

list_of_nan_ops = [
    "nansum",
    "nanprod",
]


def get_func(name):
    return wrapper_function("a", f"np.{name}(a)", globals())


@pytest.fixture(params=list_of_unary_ops + list_of_nan_ops)
def unary_op(request):
    fn = get_func(request.param)
    return fn, request.param


@pytest.fixture(params=list_of_nan_ops)
def unary_nan_op(request):
    fn = get_func(request.param)
    return fn, request.param


def test_unary_ops(filter_str, unary_op, input_array, get_shape, capfd):
    if skip_test(filter_str):
        pytest.skip()

    a = input_array
    a = np.reshape(a, get_shape)
    op, name = unary_op
    if (name == "cumprod" or name == "cumsum") and (
        filter_str == "opencl:cpu:0" or is_gen12(filter_str)
    ):
        pytest.skip()
    actual = np.empty(shape=a.shape, dtype=a.dtype)
    expected = np.empty(shape=a.shape, dtype=a.dtype)

    f = njit(op)
    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), dpnp_debug():
        actual = f(a)
        captured = capfd.readouterr()
        assert "dpnp implementation" in captured.out

    expected = op(a)
    max_abs_err = np.sum(actual - expected)
    assert max_abs_err < 1e-4


def test_unary_nan_ops(filter_str, unary_nan_op, input_nan_array, get_shape, capfd):
    if skip_test(filter_str):
        pytest.skip()

    a = input_nan_array
    a = np.reshape(a, get_shape)
    op, name = unary_nan_op
    actual = np.empty(shape=a.shape, dtype=a.dtype)
    expected = np.empty(shape=a.shape, dtype=a.dtype)

    if name == "nansum" and is_gen12(filter_str):
        pytest.skip()

    f = njit(op)
    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), dpnp_debug():
        actual = f(a)
        captured = capfd.readouterr()
        assert "dpnp implementation" in captured.out

    expected = op(a)
    max_abs_err = np.sum(actual - expected)
    assert max_abs_err < 1e-4



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numba import types
from numba.core.extending import overload, register_jitable
from numba.core.typing import signature

import numba_dppy.dpnp_glue as dpnp_lowering
import numba_dppy.dpnp_glue.dpnpimpl as dpnp_ext
from numba_dppy import dpctl_functions

from . import stubs


@overload(stubs.dpnp.repeat)
def dpnp_repeat_impl(a, repeats):
    name = "repeat"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.6.2/dpnp/backend/kernels/dpnp_krnl_manipulation.cpp#L46
    Function declaration:
    void dpnp_repeat_c(const void* array1_in, void* result1, const size_t repeats, const size_t size)
    """
    sig = signature(
        ret_type,
        types.voidptr,
        types.voidptr,
        types.intp,
        types.intp,
    )
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, [a.dtype.name, "NONE"], sig)

    PRINT_DEBUG = dpnp_lowering.DEBUG

    def dpnp_impl(a, repeats):
        if a.size == 0:
            raise ValueError("Passed Empty array")

        if a.ndim >= 2:
            raise ValueError("Not supported in dpnp")

        new_size = a.size * repeats

        out = np.zeros(new_size, dtype=a.dtype)

        sycl_queue = dpctl_functions.get_current_queue()

        a_usm = dpctl_functions.malloc_shared(a.size * a.itemsize, sycl_queue)
        event = dpctl_functions.queue_memcpy(
            sycl_queue, a_usm, a.ctypes, a.size * a.itemsize
        )
        dpctl_functions.event_wait(event)
        dpctl_functions.event_delete(event)

        out_usm = dpctl_functions.malloc_shared(out.size * out.itemsize, sycl_queue)

        dpnp_func(a_usm, out_usm, repeats, a.size)

        event = dpctl_functions.queue_memcpy(
            sycl_queue, out.ctypes, out_usm, out.size * out.itemsize
        )
        dpctl_functions.event_wait(event)
        dpctl_functions.event_delete(event)

        dpctl_functions.free_with_queue(a_usm, sycl_queue)
        dpctl_functions.free_with_queue(out_usm, sycl_queue)

        dpnp_ext._dummy_liveness_func([a.size, out.size])

        if PRINT_DEBUG:
            print("dpnp implementation")

        return out

    return dpnp_impl



################################################################################
#                                 Numba-DPPY
#
# Copyright 2020-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import dpctl
import numpy as np
import pytest
from numba import njit

import numba_dppy as dppy
from numba_dppy.tests._helper import assert_auto_offloading, skip_test

list_of_filter_strs = [
    "opencl:gpu:0",
    "level_zero:gpu:0",
    "opencl:cpu:0",
]


@pytest.fixture(params=list_of_filter_strs)
def filter_str(request):
    return request.param


list_of_binary_ops = [
    "greater",
    "greater_equal",
    "less",
    "less_equal",
    "not_equal",
    "equal",
    "logical_and",
    "logical_or",
    "logical_xor",
]


@pytest.fixture(params=list_of_binary_ops)
def binary_op(request):
    return request.param


list_of_unary_ops = [
    "isinf",
    "isfinite",
    "isnan",
]


@pytest.fixture(params=list_of_unary_ops)
def unary_op(request):
    return request.param


list_of_dtypes = [
    np.int32,
    np.int64,
    np.float32,
    np.float64,
]


@pytest.fixture(params=list_of_dtypes)
def input_arrays(request):
    # The size of input and out arrays to be used
    N = 2048
    a = np.array(np.random.random(N), request.param)
    b = np.array(np.random.random(N), request.param)
    return a, b


def test_binary_ops(filter_str, binary_op, input_arrays):
    if skip_test(filter_str):
        pytest.skip()

    a, b = input_arrays
    binop = getattr(np, binary_op)
    actual = np.empty(shape=a.shape, dtype=a.dtype)
    expected = np.empty(shape=a.shape, dtype=a.dtype)

    @njit
    def f(a, b):
        return binop(a, b)

    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), assert_auto_offloading():
        actual = f(a, b)

    expected = binop(a, b)
    np.testing.assert_allclose(actual, expected, rtol=1e-5, atol=0)


def test_unary_ops(filter_str, unary_op, input_arrays):
    if skip_test(filter_str):
        pytest.skip()

    a = input_arrays[0]
    uop = getattr(np, unary_op)
    actual = np.empty(shape=a.shape, dtype=a.dtype)
    expected = np.empty(shape=a.shape, dtype=a.dtype)

    @njit
    def f(a):
        return uop(a)

    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), assert_auto_offloading():
        actual = f(a)

    expected = uop(a)
    np.testing.assert_allclose(actual, expected, rtol=1e-5, atol=0)



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines constants that are used in other modules.

"""
from enum import Enum

__all__ = ["address_space", "calling_conv"]


class address_space:
    """The address space values supported by numba-dppy.

    ==================   ============
    Address space        Value
    ==================   ============
    PRIVATE              0
    GLOBAL               1
    CONSTANT             2
    LOCAL                3
    GENERIC              4
    ==================   ============

    """

    PRIVATE = 0
    GLOBAL = 1
    CONSTANT = 2
    LOCAL = 3
    GENERIC = 4


class calling_conv:
    CC_SPIR_KERNEL = "spir_kernel"
    CC_SPIR_FUNC = "spir_func"



#! /usr/bin/env python
# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import dpctl
import pytest
from numba.core import types

import numba_dppy as dppy
from numba_dppy import compiler
from numba_dppy.tests._helper import override_config
from numba_dppy.utils import npytypes_array_to_dppy_array

debug_options = [True, False]


@pytest.fixture(params=debug_options)
def debug_option(request):
    return request.param


def get_kernel_ir(sycl_queue, fn, sig, debug=None):
    kernel = compiler.compile_kernel(sycl_queue, fn.py_func, sig, None, debug=debug)
    return kernel.assembly


def make_check(ir, val_to_search):
    """
    Check the compiled assembly for debuginfo.
    """

    m = re.search(val_to_search, ir, re.I)
    got = m is not None
    return got


def test_debug_flag_generates_ir_with_debuginfo(debug_option):
    """
    Check debug info is emitting to IR if debug parameter is set to True
    """

    @dppy.kernel
    def foo(x):
        x = 1  # noqa

    sycl_queue = dpctl.get_current_queue()
    sig = (types.int32,)

    kernel_ir = get_kernel_ir(sycl_queue, foo, sig, debug=debug_option)

    tag = "!dbg"

    if debug_option:
        assert tag in kernel_ir
    else:
        assert tag not in kernel_ir


def test_debug_info_locals_vars_on_no_opt():
    """
    Check llvm debug tag DILocalVariable is emitting to IR for all variables if debug parameter is set to True
    and optimization is O0
    """

    @dppy.kernel
    def foo(var_a, var_b, var_c):
        i = dppy.get_global_id(0)
        var_c[i] = var_a[i] + var_b[i]

    ir_tags = [
        '!DILocalVariable(name: "var_a"',
        '!DILocalVariable(name: "var_b"',
        '!DILocalVariable(name: "var_c"',
        '!DILocalVariable(name: "i"',
    ]

    sycl_queue = dpctl.get_current_queue()
    sig = (
        npytypes_array_to_dppy_array(types.float32[:]),
        npytypes_array_to_dppy_array(types.float32[:]),
        npytypes_array_to_dppy_array(types.float32[:]),
    )

    with override_config("OPT", 0):
        kernel_ir = get_kernel_ir(sycl_queue, foo, sig, debug=True)

    for tag in ir_tags:
        assert tag in kernel_ir


def test_debug_kernel_local_vars_in_ir():
    """
    Check llvm debug tag DILocalVariable is emitting to IR for variables created in kernel
    """

    @dppy.kernel
    def foo(arr):
        index = dppy.get_global_id(0)
        local_d = 9 * 99 + 5
        arr[index] = local_d + 100

    ir_tags = ['!DILocalVariable(name: "index"', '!DILocalVariable(name: "local_d"']

    sycl_queue = dpctl.get_current_queue()
    sig = (npytypes_array_to_dppy_array(types.float32[:]),)

    kernel_ir = get_kernel_ir(sycl_queue, foo, sig, debug=True)

    for tag in ir_tags:
        assert tag in kernel_ir


def test_debug_flag_generates_ir_with_debuginfo_for_func(debug_option):
    """
    Check debug info is emitting to IR if debug parameter is set to True
    """

    @dppy.func(debug=debug_option)
    def func_sum(a, b):
        result = a + b
        return result

    @dppy.kernel(debug=debug_option)
    def data_parallel_sum(a, b, c):
        i = dppy.get_global_id(0)
        c[i] = func_sum(a[i], b[i])

    ir_tags = [
        r'\!DISubprogram\(name: ".*func_sum"',
        r'\!DISubprogram\(name: ".*data_parallel_sum"',
    ]

    sycl_queue = dpctl.get_current_queue()
    sig = (
        npytypes_array_to_dppy_array(types.float32[:]),
        npytypes_array_to_dppy_array(types.float32[:]),
        npytypes_array_to_dppy_array(types.float32[:]),
    )

    kernel_ir = get_kernel_ir(sycl_queue, data_parallel_sum, sig, debug=debug_option)

    for tag in ir_tags:
        assert debug_option == make_check(kernel_ir, tag)


def test_env_var_generates_ir_with_debuginfo_for_func(debug_option):
    """
    Check debug info is emitting to IR if NUMBA_DPPY_DEBUGINFO is set to 1
    """

    @dppy.func
    def func_sum(a, b):
        result = a + b
        return result

    @dppy.kernel
    def data_parallel_sum(a, b, c):
        i = dppy.get_global_id(0)
        c[i] = func_sum(a[i], b[i])

    ir_tags = [
        r'\!DISubprogram\(name: ".*func_sum"',
        r'\!DISubprogram\(name: ".*data_parallel_sum"',
    ]

    sycl_queue = dpctl.get_current_queue()
    sig = (
        npytypes_array_to_dppy_array(types.float32[:]),
        npytypes_array_to_dppy_array(types.float32[:]),
        npytypes_array_to_dppy_array(types.float32[:]),
    )

    with override_config("DEBUGINFO_DEFAULT", int(debug_option)):
        kernel_ir = get_kernel_ir(sycl_queue, data_parallel_sum, sig)

    for tag in ir_tags:
        assert debug_option == make_check(kernel_ir, tag)


def test_debuginfo_DISubprogram_linkageName():
    @dppy.kernel
    def func(a, b):
        i = dppy.get_global_id(0)
        b[i] = a[i]

    ir_tags = [
        r'\!DISubprogram\(.*linkageName: ".*e4func.*"',  # e4func is func(), e8func$241 is func$1()
    ]

    sycl_queue = dpctl.get_current_queue()
    sig = (
        npytypes_array_to_dppy_array(types.float32[:]),
        npytypes_array_to_dppy_array(types.float32[:]),
    )

    kernel_ir = get_kernel_ir(sycl_queue, func, sig, debug=True)

    for tag in ir_tags:
        assert make_check(kernel_ir, tag)


def test_debuginfo_DICompileUnit_language_and_producer():
    @dppy.kernel
    def func(a, b):
        i = dppy.get_global_id(0)
        b[i] = a[i]

    ir_tags = [
        r"\!DICompileUnit\(language: DW_LANG_C_plus_plus,",
        r'\!DICompileUnit\(.*producer: "numba-dppy"',
    ]

    sycl_queue = dpctl.get_current_queue()
    sig = (
        npytypes_array_to_dppy_array(types.float32[:]),
        npytypes_array_to_dppy_array(types.float32[:]),
    )

    kernel_ir = get_kernel_ir(sycl_queue, func, sig, debug=True)

    for tag in ir_tags:
        assert make_check(kernel_ir, tag)



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dpctl
import numpy as np
import pytest

import numba_dppy
from numba_dppy.tests._helper import skip_test


@numba_dppy.kernel
def sum(a, b, c):
    i = numba_dppy.get_global_id(0)
    c[i] = a[i] + b[i]


def test_strided_array_kernel(offload_device):
    if skip_test(offload_device):
        pytest.skip()

    global_size = 606
    a = np.arange(global_size * 2, dtype="i4")[::2]
    b = np.arange(global_size, dtype="i4")[::-1]
    got = np.zeros(global_size, dtype="i4")

    expected = a + b

    with dpctl.device_context(offload_device):
        sum[global_size, numba_dppy.DEFAULT_LOCAL_SIZE](a, b, got)

    assert np.array_equal(expected, got)



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numba import types
from numba.core.typing import signature


def dpctl_get_current_queue():
    ret_type = types.voidptr
    sig = signature(ret_type)
    return types.ExternalFunction("DPCTLQueueMgr_GetCurrentQueue", sig)


def dpctl_malloc_shared():
    ret_type = types.voidptr
    sig = signature(ret_type, types.int64, types.voidptr)
    return types.ExternalFunction("DPCTLmalloc_shared", sig)


def dpctl_queue_memcpy():
    ret_type = types.voidptr
    sig = signature(ret_type, types.voidptr, types.voidptr, types.voidptr, types.int64)
    return types.ExternalFunction("DPCTLQueue_Memcpy", sig)


def dpctl_event_wait():
    ret_type = types.voidptr
    sig = signature(ret_type, types.voidptr)
    return types.ExternalFunction("DPCTLEvent_Wait", sig)


def dpctl_event_delete():
    ret_type = types.void
    sig = signature(ret_type, types.voidptr)
    return types.ExternalFunction("DPCTLEvent_Delete", sig)


def dpctl_free_with_queue():
    ret_type = types.void
    sig = signature(ret_type, types.voidptr, types.voidptr)
    return types.ExternalFunction("DPCTLfree_with_queue", sig)


def dpctl_queue_wait():
    ret_type = types.void
    sig = signature(ret_type, types.voidptr)
    return types.ExternalFunction("DPCTLQueue_Wait", sig)


get_current_queue = dpctl_get_current_queue()
malloc_shared = dpctl_malloc_shared()
queue_memcpy = dpctl_queue_memcpy()
free_with_queue = dpctl_free_with_queue()
event_wait = dpctl_event_wait()
event_delete = dpctl_event_delete()
queue_wait = dpctl_queue_wait()



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numba.parfors.parfor import ParforDiagnostics, _termwidth, print_wrapped


class ExtendedParforDiagnostics(ParforDiagnostics):
    def __init__(self):
        ParforDiagnostics.__init__(self)
        self.extra_info = {}

    def dump(self, level=1):
        if level == 0:
            level = 1
        super().dump(level)

        if self.extra_info:
            parfors_simple = self.get_parfors_simple(False)
            all_lines = self.get_all_lines(parfors_simple)
            print(" Auto-offloading ".center(_termwidth, "-"))
            self.print_auto_offloading(all_lines)
            if "kernel" in self.extra_info.keys():
                print_wrapped("Device - '%s'" % self.extra_info["kernel"])
            print(_termwidth * "-")

    def print_auto_offloading(self, lines):
        # Code partially borrowed from https://github.com/IntelPython/numba/blob/97fe221b3704bd17567b57ea47f4fc6604476cf9/numba/parfors/parfor.py#L982
        sword = "+--"
        fac = len(sword)
        fadj, froots = self.compute_graph_info(self.fusion_info)
        nadj, _nroots = self.compute_graph_info(self.nested_fusion_info)

        if len(fadj) > len(nadj):
            lim = len(fadj)
            tmp = nadj
        else:
            lim = len(nadj)
            tmp = fadj
        for x in range(len(tmp), lim):
            tmp.append([])

        summary = dict()
        # region : {fused, serialized}

        def print_nest(fadj_, nadj_, theroot, reported, region_id):
            def print_g(fadj_, nadj_, nroot, depth):
                for k in nadj_[nroot]:
                    msg = fac * depth * " " + "%s%s %s" % (sword, k, "(serial")
                    if nadj_[k] == []:
                        fused = []
                        if fadj_[k] != [] and k not in reported:
                            fused = sorted(self.reachable_nodes(fadj_, k))
                            msg += ", fused with loop(s): "
                            msg += ", ".join([str(x) for x in fused])
                        msg += ")"
                        reported.append(k)
                        print_wrapped(msg)
                        summary[region_id]["fused"] += len(fused)
                    else:
                        print_wrapped(msg + ")")
                        print_g(fadj_, nadj_, k, depth + 1)
                    summary[region_id]["serialized"] += 1

            if nadj_[theroot] != []:
                print_wrapped("Parallel region %s:" % region_id)
                print_wrapped("%s%s %s" % (sword, theroot, "(parallel)"))
                summary[region_id] = {"root": theroot, "fused": 0, "serialized": 0}
                print_g(fadj_, nadj_, theroot, 1)
                print("\n")
                region_id = region_id + 1
            return region_id

        def print_fuse(ty, pf_id, adj, depth, region_id):
            print_wrapped("Parallel region %s:" % region_id)
            msg = fac * depth * " " + "%s%s %s" % (sword, pf_id, "(parallel")
            fused = []
            if adj[pf_id] != []:
                fused = sorted(self.reachable_nodes(adj, pf_id))
                msg += ", fused with loop(s): "
                msg += ", ".join([str(x) for x in fused])

            summary[region_id] = {"root": pf_id, "fused": len(fused), "serialized": 0}
            msg += ")"
            print_wrapped(msg)
            extra_info = self.extra_info.get(str(region_id))
            if extra_info:
                print_wrapped("Device - '%s'" % extra_info)
            region_id = region_id + 1
            return region_id

        # Walk the parfors by src line and print optimised structure
        region_id = 0
        reported = []
        for line, info in sorted(lines.items()):
            opt_ty, pf_id, adj = info
            if opt_ty == "fuse":
                if pf_id not in reported:
                    region_id = print_fuse("f", pf_id, adj, 0, region_id)
            elif opt_ty == "nest":
                region_id = print_nest(fadj, nadj, pf_id, reported, region_id)
            else:
                assert 0

        # print the summary of the fuse/serialize rewrite
        if summary:
            for k, v in sorted(summary.items()):
                msg = "\n \nParallel region %s (loop #%s) had %s " "loop(s) fused"
                root = v["root"]
                fused = v["fused"]
                serialized = v["serialized"]
                if serialized != 0:
                    msg += (
                        " and %s loop(s) "
                        "serialized as part of the larger "
                        "parallel loop (#%s)."
                    )
                    print_wrapped(msg % (k, root, fused, serialized, root))
                else:
                    msg += "."
                    print_wrapped(msg % (k, root, fused))
        else:
            print_wrapped("Parallel structure is already optimal.")



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import numpy as np
from llvmlite import binding as ll
from llvmlite import ir as llvmir
from llvmlite.llvmpy import core as lc
from numba import typeof
from numba.core import cgutils, datamodel, types, typing, utils
from numba.core.base import BaseContext
from numba.core.callconv import MinimalCallConv
from numba.core.registry import cpu_target
from numba.core.target_extension import GPU, target_registry
from numba.core.utils import cached_property

from numba_dppy.dppy_array_type import DPPYArray, DPPYArrayModel
from numba_dppy.utils import address_space, calling_conv, npytypes_array_to_dppy_array

from . import codegen

CC_SPIR_KERNEL = "spir_kernel"
CC_SPIR_FUNC = "spir_func"
VALID_CHARS = re.compile(r"[^a-z0-9]", re.I)
LINK_ATOMIC = 111
LLVM_SPIRV_ARGS = 112


class DPPYTypingContext(typing.BaseContext):
    """A numba_dppy-specific typing context inheriting Numba's ``BaseContext``.

    :class:`DPPYTypingContext` is a customized typing context that inherits from
    Numba's ``typing.BaseContext`` class. We add two specific functionalities
    to the basic Numba typing context features: An overridden
    :func:`resolve_argument_type` that changes all ``npytypes.Array`` to
    numba-dppy's :class:`dppy_array_type.DppyArray`. An overridden
    :func:`load_additional_registries` that registers OpenCL math and other
    functions to the typing context.

    """

    def resolve_argument_type(self, val):
        """Return the Numba type of a Python value used as a function argument.

        Overrides the implementation of ``numba.core.typing.BaseContext`` to
        handle the special case of ``numba.core.types.npytypes.Array``. Whenever
        a NumPy ndarray argument is encountered as an argument to a ``kernel``
        function, it is converted to a ``DPPYArray`` type.

        Args:
            val : A Python value that is passed as an argument to a ``kernel``
                  function.

        Returns: The Numba type corresponding to the Python value.

        Raises:
            ValueError: If the type of the Python value is not supported.

        """
        if type(typeof(val)) is types.npytypes.Array:
            # Convert npytypes.Array to DPPYArray
            return npytypes_array_to_dppy_array(typeof(val))
        else:
            return super().resolve_argument_type(val)

    def load_additional_registries(self):
        """Register the OpenCL API and math and other functions."""
        from numba.core.typing import cmathdecl, npydecl

        from .ocl import mathdecl, ocldecl

        self.install_registry(ocldecl.registry)
        self.install_registry(mathdecl.registry)
        self.install_registry(cmathdecl.registry)
        self.install_registry(npydecl.registry)


class GenericPointerModel(datamodel.PrimitiveModel):
    def __init__(self, dmm, fe_type):
        adrsp = (
            fe_type.addrspace if fe_type.addrspace is not None else address_space.GLOBAL
        )
        be_type = dmm.lookup(fe_type.dtype).get_data_type().as_pointer(adrsp)
        super(GenericPointerModel, self).__init__(dmm, fe_type, be_type)


def _init_data_model_manager():
    dmm = datamodel.default_manager.copy()
    dmm.register(types.CPointer, GenericPointerModel)
    dmm.register(DPPYArray, DPPYArrayModel)
    return dmm


spirv_data_model_manager = _init_data_model_manager()


class SyclDevice(GPU):
    """Mark the hardware target as SYCL Device."""

    pass


DPPY_TARGET_NAME = "SyclDevice"

target_registry[DPPY_TARGET_NAME] = SyclDevice

import numba_dppy.dppy_offload_dispatcher


class DPPYTargetContext(BaseContext):
    """A numba_dppy-specific target context inheriting Numba's ``BaseContext``.

    :class:`DPPYTargetContext` is a customized target context that inherits
    from Numba's ``numba.core.base.BaseContext`` class. The class defines
    helper functions to mark LLVM functions as SPIR-V kernels. The class also
    registers OpenCL math and API functions, helper functions for inserting
    LLVM address space cast instructions, and other functionalities used by
    numba_dppy's compiler passes.

    """

    implement_powi_as_math_call = True

    def _gen_arg_addrspace_md(self, fn):
        """Generate kernel_arg_addr_space metadata."""
        mod = fn.module
        fnty = fn.type.pointee
        codes = []

        for a in fnty.args:
            if cgutils.is_pointer(a):
                codes.append(address_space.GLOBAL)
            else:
                codes.append(address_space.PRIVATE)

        consts = [lc.Constant.int(lc.Type.int(), x) for x in codes]
        name = lc.MetaDataString.get(mod, "kernel_arg_addr_space")
        return lc.MetaData.get(mod, [name] + consts)

    def _gen_arg_access_qual_md(self, fn):
        """Generate kernel_arg_access_qual metadata."""
        mod = fn.module
        consts = [lc.MetaDataString.get(mod, "none")] * len(fn.args)
        name = lc.MetaDataString.get(mod, "kernel_arg_access_qual")
        return lc.MetaData.get(mod, [name] + consts)

    def _gen_arg_type(self, fn):
        """Generate kernel_arg_type metadata."""
        mod = fn.module
        fnty = fn.type.pointee
        consts = [lc.MetaDataString.get(mod, str(a)) for a in fnty.args]
        name = lc.MetaDataString.get(mod, "kernel_arg_type")
        return lc.MetaData.get(mod, [name] + consts)

    def _gen_arg_type_qual(self, fn):
        """Generate kernel_arg_type_qual metadata."""
        mod = fn.module
        fnty = fn.type.pointee
        consts = [lc.MetaDataString.get(mod, "") for _ in fnty.args]
        name = lc.MetaDataString.get(mod, "kernel_arg_type_qual")
        return lc.MetaData.get(mod, [name] + consts)

    def _gen_arg_base_type(self, fn):
        """Generate kernel_arg_base_type metadata."""
        mod = fn.module
        fnty = fn.type.pointee
        consts = [lc.MetaDataString.get(mod, str(a)) for a in fnty.args]
        name = lc.MetaDataString.get(mod, "kernel_arg_base_type")
        return lc.MetaData.get(mod, [name] + consts)

    def _finalize_wrapper_module(self, fn):
        """Add metadata and calling convention to the wrapper function.

        The helper function adds function metadata to the wrapper function and
        also module level metadata to the LLVM module containing the wrapper.
        We also make sure the wrapper function has ``spir_kernel`` calling
        convention, without which the function cannot be used as a kernel.

        Args:
            fn: LLVM function representing the "kernel" wrapper function.

        """
        mod = fn.module
        # Set norecurse
        fn.attributes.add("norecurse")
        # Set SPIR kernel calling convention
        fn.calling_convention = CC_SPIR_KERNEL

        # Mark kernels
        ocl_kernels = mod.get_or_insert_named_metadata("opencl.kernels")
        ocl_kernels.add(
            lc.MetaData.get(
                mod,
                [
                    fn,
                    self._gen_arg_addrspace_md(fn),
                    self._gen_arg_access_qual_md(fn),
                    self._gen_arg_type(fn),
                    self._gen_arg_type_qual(fn),
                    self._gen_arg_base_type(fn),
                ],
            )
        )

        # Other metadata
        empty_md = lc.MetaData.get(mod, ())
        others = [
            "opencl.used.extensions",
            "opencl.used.optional.core.features",
            "opencl.compiler.options",
        ]

        for name in others:
            nmd = mod.get_or_insert_named_metadata(name)
            if not nmd.operands:
                nmd.add(empty_md)

    def _generate_kernel_wrapper(self, func, argtypes):
        module = func.module
        arginfo = self.get_arg_packer(argtypes)
        wrapperfnty = lc.Type.function(lc.Type.void(), arginfo.argument_types)
        wrapper_module = self.create_module("dppy.kernel.wrapper")
        wrappername = "dppyPy_{name}".format(name=func.name)
        argtys = list(arginfo.argument_types)
        fnty = lc.Type.function(
            lc.Type.int(),
            [self.call_conv.get_return_type(types.pyobject)] + argtys,
        )
        func = wrapper_module.add_function(fnty, name=func.name)
        func.calling_convention = CC_SPIR_FUNC
        wrapper = wrapper_module.add_function(wrapperfnty, name=wrappername)
        builder = lc.Builder(wrapper.append_basic_block(""))

        callargs = arginfo.from_arguments(builder, wrapper.args)

        # XXX handle error status
        status, _ = self.call_conv.call_function(
            builder, func, types.void, argtypes, callargs
        )
        builder.ret_void()

        self._finalize_wrapper_module(wrapper)

        # Link the spir_func module to the wrapper module
        module.link_in(ll.parse_assembly(str(wrapper_module)))
        # Make sure the spir_func has internal linkage to be inlinable.
        func.linkage = "internal"
        wrapper = module.get_function(wrapper.name)
        module.get_function(func.name).linkage = "internal"
        return wrapper

    def __init__(self, typingctx, target=DPPY_TARGET_NAME):
        super().__init__(typingctx, target)

    def init(self):
        self._internal_codegen = codegen.JITSPIRVCodegen("numba_dppy.jit")
        self._target_data = ll.create_target_data(
            codegen.SPIR_DATA_LAYOUT[utils.MACHINE_BITS]
        )
        # Override data model manager to SPIR model
        import numba.cpython.unicode

        self.data_model_manager = _init_data_model_manager()
        self.extra_compile_options = dict()

        import copy

        from numba.np.ufunc_db import _lazy_init_db

        _lazy_init_db()
        from numba.np.ufunc_db import _ufunc_db as ufunc_db

        self.ufunc_db = copy.deepcopy(ufunc_db)
        self.cpu_context = cpu_target.target_context

    # Overrides
    def create_module(self, name):
        return self._internal_codegen._create_empty_module(name)

    def replace_numpy_ufunc_with_opencl_supported_functions(self):
        from numba_dppy.ocl.mathimpl import lower_ocl_impl, sig_mapper

        ufuncs = [
            ("fabs", np.fabs),
            ("exp", np.exp),
            ("log", np.log),
            ("log10", np.log10),
            ("expm1", np.expm1),
            ("log1p", np.log1p),
            ("sqrt", np.sqrt),
            ("sin", np.sin),
            ("cos", np.cos),
            ("tan", np.tan),
            ("asin", np.arcsin),
            ("acos", np.arccos),
            ("atan", np.arctan),
            ("atan2", np.arctan2),
            ("sinh", np.sinh),
            ("cosh", np.cosh),
            ("tanh", np.tanh),
            ("asinh", np.arcsinh),
            ("acosh", np.arccosh),
            ("atanh", np.arctanh),
            ("ldexp", np.ldexp),
            ("floor", np.floor),
            ("ceil", np.ceil),
            ("trunc", np.trunc),
            ("hypot", np.hypot),
            ("exp2", np.exp2),
            ("log2", np.log2),
        ]

        for name, ufunc in ufuncs:
            for sig in self.ufunc_db[ufunc].keys():
                if sig in sig_mapper and (name, sig_mapper[sig]) in lower_ocl_impl:
                    self.ufunc_db[ufunc][sig] = lower_ocl_impl[(name, sig_mapper[sig])]

    def load_additional_registries(self):
        """Register OpenCL functions into numba-dppy's target context.

        To make sure we are calling supported OpenCL math functions, we
        replace some of NUMBA's NumPy ufunc with OpenCL versions of those
        functions. The replacement is done after the OpenCL functions have
        been registered into the target context.

        """
        from numba.np import npyimpl

        from . import printimpl
        from .ocl import mathimpl, oclimpl

        self.insert_func_defn(oclimpl.registry.functions)
        self.insert_func_defn(mathimpl.registry.functions)
        self.insert_func_defn(npyimpl.registry.functions)
        self.install_registry(printimpl.registry)
        # Replace NumPy functions with their OpenCL versions.
        self.replace_numpy_ufunc_with_opencl_supported_functions()

    @cached_property
    def call_conv(self):
        return DPPYCallConv(self)

    def codegen(self):
        return self._internal_codegen

    @property
    def target_data(self):
        return self._target_data

    def mangler(self, name, argtypes):
        def repl(m):
            ch = m.group(0)
            return "_%X_" % ord(ch)

        qualified = name + "." + ".".join(str(a) for a in argtypes)
        mangled = VALID_CHARS.sub(repl, qualified)
        return "dppy_py_devfn_" + mangled

    def prepare_ocl_kernel(self, func, argtypes):
        module = func.module
        func.linkage = "linkonce_odr"
        module.data_layout = codegen.SPIR_DATA_LAYOUT[self.address_size]
        wrapper = self._generate_kernel_wrapper(func, argtypes)
        return wrapper

    def mark_ocl_device(self, func):
        # Adapt to SPIR
        func.calling_convention = CC_SPIR_FUNC
        func.linkage = "linkonce_odr"
        return func

    def declare_function(self, module, fndesc):
        """Create the LLVM function from a ``numba_dppy.kernel`` decorated
        function.

        Args:
            module (llvmlite.llvmpy.core.Module) : The LLVM module into which
                the kernel function will be inserted.
            fndesc (numba.core.funcdesc.PythonFunctionDescriptor) : The
                signature of the function.

        Returns:
            llvmlite.ir.values.Function: The reference to the LLVM Function
                that was inserted into the module.

        """
        fnty = self.call_conv.get_function_type(fndesc.restype, fndesc.argtypes)
        fn = module.get_or_insert_function(fnty, name=fndesc.mangled_name)
        if not self.enable_debuginfo:
            fn.attributes.add("alwaysinline")
        ret = super(DPPYTargetContext, self).declare_function(module, fndesc)
        ret.calling_convention = calling_conv.CC_SPIR_FUNC
        return ret

    def insert_const_string(self, mod, string):
        """Create a global string from the passed in string argument and return
        a void* in the GENERIC address space pointing to that string.

        Args:
            mod: LLVM module where the global string value is to be inserted.
            string: A Python string that will be converted to a global constant
                    string and inserted into the module.

        Returns: A LLVM Constant pointing to the global string value inserted
                 into the module.

        """
        text = cgutils.make_bytearray(string.encode("utf-8") + b"\x00")

        name = "$".join(["__conststring__", self.mangler(string, ["str"])])

        # Try to reuse existing global
        gv = mod.globals.get(name)
        if gv is None:
            # Not defined yet
            gv = cgutils.add_global_variable(
                mod, text.type, name=name, addrspace=address_space.GENERIC
            )
            gv.linkage = "internal"
            gv.global_constant = True
            gv.initializer = text

        # Cast to a i8* pointer
        charty = gv.type.pointee.element
        return gv.bitcast(charty.as_pointer(address_space.GENERIC))

    def addrspacecast(self, builder, src, addrspace):
        """Insert an LLVM addressspace cast instruction into the module.

        FIXME: Move this function into utils.

        """
        ptras = llvmir.PointerType(src.type.pointee, addrspace=addrspace)
        return builder.addrspacecast(src, ptras)

    # Overrides
    def get_ufunc_info(self, ufunc_key):
        return self.ufunc_db[ufunc_key]


class DPPYCallConv(MinimalCallConv):
    """Custom calling convention class used by numba-dppy.

    Numba-dppy's calling convention derives from
    :class:`numba.core.callconv import MinimalCallConv`. The
    :class:`DPPYCallConv` overriddes :func:`call_function`.

    """

    def call_function(self, builder, callee, resty, argtys, args, env=None):
        """Call the Numba-compiled *callee*."""
        assert env is None
        retty = callee.args[0].type.pointee
        retvaltmp = cgutils.alloca_once(builder, retty)
        # initialize return value
        builder.store(cgutils.get_null_value(retty), retvaltmp)
        arginfo = self.context.get_arg_packer(argtys)
        args = arginfo.as_arguments(builder, args)
        realargs = [retvaltmp] + list(args)
        code = builder.call(callee, realargs)
        status = self._get_return_status(builder, code)
        retval = builder.load(retvaltmp)
        out = self.context.get_returned_value(builder, resty, retval)
        return status, out



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import llvmlite.ir as lir
import llvmlite.llvmpy.core as lc
from numba.core import cgutils, types


class LLVMTypes:
    """
    A helper class to get LLVM Values for integer C types.
    """

    byte_t = lc.Type.int(8)
    byte_ptr_t = lc.Type.pointer(byte_t)
    byte_ptr_ptr_t = lc.Type.pointer(byte_ptr_t)
    int32_t = lc.Type.int(32)
    int32_ptr_t = lc.Type.pointer(int32_t)
    int64_t = lc.Type.int(64)
    int64_ptr_t = lc.Type.pointer(int64_t)
    void_t = lir.VoidType()


def get_llvm_type(context, type):
    """Returns the LLVM Value corresponsing to a Numba type.

    Args:
        context: The LLVM context or the execution state of the current IR
                 generator.
        type: A Numba type object.

    Returns: An Python object wrapping an LLVM Value corresponding to the
             specified Numba type.
    """
    return context.get_value_type(type)


def get_llvm_ptr_type(type):
    """Returns an LLVM pointer type for a give LLVM type object.

    Args:
        type: An LLVM type for which we need the corresponding pointer type.

    Returns: An LLVM pointer type object corresponding to the input LLVM type.
    """
    return lc.Type.pointer(type)


def create_null_ptr(builder, context):
    """
    Allocates a new LLVM Value storing a ``void*`` and returns the Value to
    caller.

    Args:
        builder: The LLVM IR builder to be used for code generation.
        context: The LLVM IR builder context.

    Returns: An LLVM value storing a null pointer
    """
    null_ptr = cgutils.alloca_once(
        builder=builder,
        ty=context.get_value_type(types.voidptr),
        size=context.get_constant(types.uintp, 1),
    )
    builder.store(
        builder.inttoptr(
            context.get_constant(types.uintp, 0),
            get_llvm_type(context=context, type=types.voidptr),
        ),
        null_ptr,
    )
    return null_ptr


def get_zero(context):
    """Returns an LLVM Constant storing a 64 bit representation for zero.

    Args:
        context: The LLVM IR builder context.

    Returns: An LLVM Contant Value storing zero.
    """
    return context.get_constant(types.uintp, 0)


def get_one(context):
    """Returns an LLVM Constant storing a 64 bit representation for one.

    Args:
        context: The LLVM IR builder context.

    Returns: An LLVM Contant Value storing one.
    """
    return context.get_constant(types.uintp, 1)



################################################################################
#                                 Numba-DPPY
#
# Copyright 2020-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import dpctl
import numpy as np
import pytest
from numba import njit

import numba_dppy as dppy
from numba_dppy.tests._helper import dpnp_debug

from ._helper import wrapper_function
from .dpnp_skip_test import dpnp_skip_test as skip_test

list_of_filter_strs = [
    "opencl:gpu:0",
    "level_zero:gpu:0",
    "opencl:cpu:0",
]


@pytest.fixture(params=list_of_filter_strs)
def filter_str(request):
    return request.param


list_of_dtypes = [
    np.int32,
    np.int64,
    np.float32,
    np.float64,
]


@pytest.fixture(params=list_of_dtypes)
def input_arrays(request):
    # The size of input and out arrays to be used
    N = 10
    a = np.array(np.random.random(N), request.param)
    b = np.array(np.random.random(N), request.param)
    return a, b


list_of_shape = [
    (10),
    (5, 2),
]


@pytest.fixture(params=list_of_shape)
def get_shape(request):
    return request.param


list_of_unary_ops = [
    "sort",
    "argmax",
    "argmin",
    "argsort",
]


@pytest.fixture(params=list_of_unary_ops)
def unary_op(request):
    return wrapper_function("a", f"np.{request.param}(a)", globals()), request.param


def test_unary_ops(filter_str, unary_op, input_arrays, get_shape, capfd):
    if skip_test(filter_str):
        pytest.skip()

    a = input_arrays[0]
    op, name = unary_op
    if name != "argsort" and name != "sort":
        a = np.reshape(a, get_shape)
    actual = np.empty(shape=a.shape, dtype=a.dtype)
    expected = np.empty(shape=a.shape, dtype=a.dtype)

    f = njit(op)
    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), dpnp_debug():
        actual = f(a)
        captured = capfd.readouterr()
        assert "dpnp implementation" in captured.out

    expected = op(a)
    np.testing.assert_allclose(actual, expected, rtol=1e-3, atol=0)


@pytest.mark.parametrize("kth", [0, 1], ids=["0", "1"])
@pytest.mark.parametrize(
    "array",
    [
        [3, 4, 2, 1],
        [[1, 0], [3, 0]],
        [[3, 2], [1, 6]],
        [[4, 2, 3], [3, 4, 1]],
        [[[1, -3], [3, 0]], [[5, 2], [0, 1]], [[1, 0], [0, 1]]],
        [[[[8, 2], [3, 0]], [[5, 2], [0, 1]]], [[[1, 3], [3, 1]], [[5, 2], [0, 1]]]],
    ],
    ids=[
        "[3, 4, 2, 1]",
        "[[1, 0], [3, 0]]",
        "[[3, 2], [1, 6]]",
        "[[4, 2, 3], [3, 4, 1]]",
        "[[[1, -3], [3, 0]], [[5, 2], [0, 1]], [[1, 0], [0, 1]]]",
        "[[[[8, 2], [3, 0]], [[5, 2], [0, 1]]], [[[1, 3], [3, 1]], [[5, 2], [0, 1]]]]",
    ],
)
def test_partition(array, kth, filter_str):
    if skip_test(filter_str):
        pytest.skip()

    a = np.array(array)

    def fn(a, kth):
        return np.partition(a, kth)

    f = njit(fn)
    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), dpnp_debug():
        actual = f(a, kth)

    expected = fn(a, kth)
    np.testing.assert_allclose(actual, expected, rtol=1e-3, atol=0)



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Support for interoperability."""

import dpctl.tensor as dpt


def asarray(container):
    """Convert container supported by interoperability to numba-dppy container.
    Currently used dpctl.tensor.asarray().
    """
    try:
        return dpt.asarray(container)
    except:
        pass

    # Workaround for dpnp_array if dpctl asarray() does not support it.
    try:
        from dpnp.dpnp_array import dpnp_array

        if isinstance(container, dpnp_array) and hasattr(container, "_array_obj"):
            import warnings

            warnings.warn("asarray() uses internals from dpnp.")
            return container._array_obj
    except:
        pass

    raise NotImplementedError("dpctl asarray() does not support " + type(container))



#! /usr/bin/env python
# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import dpctl
import numpy as np
from numba import njit, prange

import numba_dppy as dppy
from numba_dppy.tests._helper import assert_auto_offloading

from . import _helper


@unittest.skipUnless(_helper.has_gpu_queues(), "test only on GPU system")
class TestPrange(unittest.TestCase):
    def test_one_prange(self):
        @njit
        def f(a, b):
            for i in prange(4):
                b[i, 0] = a[i, 0] * 10

        m = 8
        n = 8
        a = np.ones((m, n))
        b = np.ones((m, n))

        device = dpctl.SyclDevice("opencl:gpu")
        with assert_auto_offloading(), dpctl.device_context(device):
            f(a, b)

        for i in range(4):
            self.assertTrue(b[i, 0] == a[i, 0] * 10)

    def test_nested_prange(self):
        @njit
        def f(a, b):
            # dimensions must be provided as scalar
            m, n = a.shape
            for i in prange(m):
                for j in prange(n):
                    b[i, j] = a[i, j] * 10

        m = 8
        n = 8
        a = np.ones((m, n))
        b = np.ones((m, n))

        device = dpctl.SyclDevice("opencl:gpu")
        with assert_auto_offloading(), dpctl.device_context(device):
            f(a, b)

        self.assertTrue(np.all(b == 10))

    def test_multiple_prange(self):
        @njit
        def f(a, b):
            # dimensions must be provided as scalar
            m, n = a.shape
            for i in prange(m):
                val = 10
                for j in prange(n):
                    b[i, j] = a[i, j] * val

            for i in prange(m):
                for j in prange(n):
                    a[i, j] = a[i, j] * 10

        m = 8
        n = 8
        a = np.ones((m, n))
        b = np.ones((m, n))

        device = dpctl.SyclDevice("opencl:gpu")
        with assert_auto_offloading(parfor_offloaded=2), dpctl.device_context(device):
            f(a, b)

        self.assertTrue(np.all(b == 10))
        self.assertTrue(np.all(a == 10))

    def test_three_prange(self):
        @njit
        def f(a, b):
            # dimensions must be provided as scalar
            m, n, o = a.shape
            for i in prange(m):
                val = 10
                for j in prange(n):
                    constant = 2
                    for k in prange(o):
                        b[i, j, k] = a[i, j, k] * (val + constant)

        m = 8
        n = 8
        o = 8
        a = np.ones((m, n, o))
        b = np.ones((m, n, o))

        device = dpctl.SyclDevice("opencl:gpu")
        with assert_auto_offloading(parfor_offloaded=1), dpctl.device_context(device):
            f(a, b)

        self.assertTrue(np.all(b == 12))

    @unittest.skip("numba-dppy issue 110")
    def test_two_consequent_prange(self):
        def prange_example():
            n = 10
            a = np.ones((n), dtype=np.float64)
            b = np.ones((n), dtype=np.float64)
            c = np.ones((n), dtype=np.float64)
            for i in prange(n // 2):
                a[i] = b[i] + c[i]

            return a

        jitted = njit(prange_example)

        device = dpctl.SyclDevice("opencl:gpu")
        with assert_auto_offloading(parfor_offloaded=2), dpctl.device_context(device):
            jitted_res = jitted()

        res = prange_example()

        np.testing.assert_equal(res, jitted_res)

    @unittest.skip("NRT required but not enabled")
    def test_2d_arrays(self):
        def prange_example():
            n = 10
            a = np.ones((n, n), dtype=np.float64)
            b = np.ones((n, n), dtype=np.float64)
            c = np.ones((n, n), dtype=np.float64)
            for i in prange(n // 2):
                a[i] = b[i] + c[i]

            return a

        jitted = njit(prange_example)

        device = dpctl.SyclDevice("opencl:gpu")
        with assert_auto_offloading(parfor_offloaded=2), dpctl.device_context(device):
            jitted_res = jitted()

        res = prange_example()

        np.testing.assert_equal(res, jitted_res)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numba import types
from numba.core.extending import overload, register_jitable
from numba.core.typing import signature

import numba_dppy
import numba_dppy.dpnp_glue as dpnp_lowering
import numba_dppy.dpnp_glue.dpnpimpl as dpnp_ext
from numba_dppy import dpctl_functions

from . import stubs


@register_jitable
def common_impl(a, b, out, dpnp_func, PRINT_DEBUG):
    if a.size == 0:
        raise ValueError("Passed Empty array")

    sycl_queue = dpctl_functions.get_current_queue()

    b_usm = dpctl_functions.malloc_shared(b.size * b.itemsize, sycl_queue)
    event = dpctl_functions.queue_memcpy(
        sycl_queue, b_usm, b.ctypes, b.size * b.itemsize
    )
    dpctl_functions.event_wait(event)
    dpctl_functions.event_delete(event)

    out_usm = dpctl_functions.malloc_shared(out.size * out.itemsize, sycl_queue)

    dpnp_func(out_usm, b_usm, a.size)

    event = dpctl_functions.queue_memcpy(
        sycl_queue, out.ctypes, out_usm, out.size * out.itemsize
    )
    dpctl_functions.event_wait(event)
    dpctl_functions.event_delete(event)

    dpctl_functions.free_with_queue(b_usm, sycl_queue)
    dpctl_functions.free_with_queue(out_usm, sycl_queue)

    dpnp_ext._dummy_liveness_func([a.size, out.size])

    if PRINT_DEBUG:
        print("dpnp implementation")


@register_jitable
def common_shape_impl(a, out, dpnp_func, PRINT_DEBUG):
    if a.size == 0:
        raise ValueError("Passed Empty array")

    sycl_queue = dpctl_functions.get_current_queue()

    a_usm = dpctl_functions.malloc_shared(a.size * a.itemsize, sycl_queue)
    event = dpctl_functions.queue_memcpy(
        sycl_queue, a_usm, a.ctypes, a.size * a.itemsize
    )
    dpctl_functions.event_wait(event)
    dpctl_functions.event_delete(event)

    out_usm = dpctl_functions.malloc_shared(out.size * out.itemsize, sycl_queue)

    dpnp_func(a_usm, out_usm, a.shapeptr, a.ndim)

    event = dpctl_functions.queue_memcpy(
        sycl_queue, out.ctypes, out_usm, out.size * out.itemsize
    )
    dpctl_functions.event_wait(event)
    dpctl_functions.event_delete(event)

    dpctl_functions.free_with_queue(a_usm, sycl_queue)
    dpctl_functions.free_with_queue(out_usm, sycl_queue)

    dpnp_ext._dummy_liveness_func([a.size, out.size])

    if PRINT_DEBUG:
        print("dpnp implementation")


@overload(stubs.dpnp.zeros_like)
def dpnp_zeros_like_impl(a, dtype=None):
    name = "zeros_like"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.5.1/dpnp/backend/kernels/dpnp_krnl_common.cpp#L224

    Function declaration:
    void dpnp_initval_c(void* result1, void* value, size_t size)

    """
    res_dtype = dtype
    if dtype == types.none or dtype is None:
        res_dtype = a.dtype
        name_dtype = res_dtype.name
    else:
        name_dtype = res_dtype.dtype.name

    sig = signature(ret_type, types.voidptr, types.voidptr, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, [name_dtype, "NONE"], sig)

    PRINT_DEBUG = dpnp_lowering.DEBUG

    def dpnp_impl(a, dtype=None):
        b = np.zeros(1, dtype=res_dtype)
        out = np.zeros(a.shape, dtype=res_dtype)
        common_impl(a, b, out, dpnp_func, PRINT_DEBUG)
        return out

    return dpnp_impl


@overload(stubs.dpnp.ones_like)
def dpnp_ones_like_impl(a, dtype=None):
    name = "ones_like"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.5.1/dpnp/backend/kernels/dpnp_krnl_common.cpp#L224

    Function declaration:
    void dpnp_initval_c(void* result1, void* value, size_t size)

    """
    res_dtype = dtype
    if dtype == types.none or dtype is None:
        res_dtype = a.dtype
        name_dtype = res_dtype.name
    else:
        name_dtype = res_dtype.dtype.name

    sig = signature(ret_type, types.voidptr, types.voidptr, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, [name_dtype, "NONE"], sig)

    PRINT_DEBUG = dpnp_lowering.DEBUG

    def dpnp_impl(a, dtype=None):
        b = np.ones(1, dtype=res_dtype)
        out = np.ones(a.shape, dtype=res_dtype)
        common_impl(a, b, out, dpnp_func, PRINT_DEBUG)
        return out

    return dpnp_impl


@overload(stubs.dpnp.full_like)
def dpnp_full_like_impl(a, b):
    name = "full_like"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.5.1/dpnp/backend/kernels/dpnp_krnl_common.cpp#L224

    Function declaration:
    void dpnp_initval_c(void* result1, void* value, size_t size)

    """
    sig = signature(ret_type, types.voidptr, types.voidptr, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, [b.dtype.name, "NONE"], sig)

    res_dtype = b.dtype
    PRINT_DEBUG = dpnp_lowering.DEBUG

    def dpnp_impl(a, b):
        out = np.ones(a.shape, dtype=res_dtype)
        common_impl(a, b, out, dpnp_func, PRINT_DEBUG)
        return out

    return dpnp_impl


# TODO: This implementation is incorrect
@overload(stubs.dpnp.full)
def dpnp_full_impl(a, b):
    name = "full"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.5.1/dpnp/backend/kernels/dpnp_krnl_arraycreation.cpp#L70

    Function declaration:
    void dpnp_full_c(void* array_in, void* result, const size_t size)

    """
    sig = signature(ret_type, types.voidptr, types.voidptr, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, [b.dtype.name, "NONE"], sig)

    res_dtype = b.dtype
    PRINT_DEBUG = dpnp_lowering.DEBUG

    def dpnp_impl(a, b):
        if a.size == 0:
            raise ValueError("Passed Empty array")

        sycl_queue = dpctl_functions.get_current_queue()

        b_usm = dpctl_functions.malloc_shared(b.size * b.itemsize, sycl_queue)
        event = dpctl_functions.queue_memcpy(
            sycl_queue, b_usm, b.ctypes, b.size * b.itemsize
        )
        dpctl_functions.event_wait(event)
        dpctl_functions.event_delete(event)

        out = np.arange(0, a.size, 1, res_dtype)
        out_usm = dpctl_functions.malloc_shared(out.size * out.itemsize, sycl_queue)

        dpnp_func(b_usm, out_usm, a.size)

        event = dpctl_functions.queue_memcpy(
            sycl_queue, out.ctypes, out_usm, out.size * out.itemsize
        )
        dpctl_functions.event_wait(event)
        dpctl_functions.event_delete(event)

        dpctl_functions.free_with_queue(b_usm, sycl_queue)
        dpctl_functions.free_with_queue(out_usm, sycl_queue)

        dpnp_ext._dummy_liveness_func([a.size, out.size])

        if PRINT_DEBUG:
            print("dpnp implementation")
        return out

    return dpnp_impl


@overload(stubs.dpnp.trace)
def dpnp_trace_impl(a):
    name = "trace"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.6.2/dpnp/backend/kernels/dpnp_krnl_arraycreation.cpp#L218

    Function declaration:
    void dpnp_trace_c(const void* array1_in, void* result1, const size_t* shape_, const size_t ndim)

    """
    sig = signature(ret_type, types.voidptr, types.voidptr, types.voidptr, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, [a.dtype.name, "NONE"], sig)

    PRINT_DEBUG = dpnp_lowering.DEBUG

    def dpnp_impl(a):
        diag_arr = numba_dppy.dpnp.diagonal(a, 0)
        out = np.zeros(diag_arr.shape[:-1], dtype=a.dtype)
        common_shape_impl(diag_arr, out, dpnp_func, PRINT_DEBUG)
        return out

    return dpnp_impl



################################################################################
#                                 Numba-DPPY
#
# Copyright 2020-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import dpctl
import numpy as np
import pytest
from numba import njit

import numba_dppy as dppy
from numba_dppy.tests._helper import dpnp_debug

from ._helper import wrapper_function
from .dpnp_skip_test import dpnp_skip_test as skip_test

# dpnp throws -30 (CL_INVALID_VALUE) when invoked with multiple kinds of
# devices at runtime, so testing for level_zero only
list_of_filter_strs = [
    # "opencl:gpu:0",
    "level_zero:gpu:0",
    # "opencl:cpu:0",
]


@pytest.fixture(params=list_of_filter_strs)
def filter_str(request):
    return request.param


list_of_size = [
    9,
    (2, 5),
    (3, 2, 4),
]

none_size = [None]


@pytest.fixture(params=list_of_size)
def unary_size(request):
    return request.param


@pytest.fixture(params=list_of_size + none_size)
def three_arg_size(request):
    return request.param


list_of_one_arg = [
    ("random_sample", 0.0, 1.0),
    ("ranf", 0.0, 1.0),
    ("sample", 0.0, 1.0),
    ("random", 0.0, 1.0),
    ("standard_exponential", 0.0, None),
    ("standard_normal", None, None),
    ("standard_cauchy", None, None),
]


@pytest.fixture(params=list_of_one_arg)
def one_arg_fn(request):
    function = wrapper_function(
        "size", f"np.random.{request.param[0]}(size)", globals()
    )
    return function, request.param


def test_one_arg_fn(filter_str, one_arg_fn, unary_size, capfd):
    if skip_test(filter_str):
        pytest.skip()

    op, params = one_arg_fn
    name, low, high = params
    f = njit(op)
    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), dpnp_debug():
        actual = f(unary_size)
        captured = capfd.readouterr()
        assert "dpnp implementation" in captured.out

        if low is not None:
            assert np.all(actual >= low)
        if high is not None:
            assert np.all(actual < high)


list_of_two_arg_fn = [
    ("chisquare", 3, 0, None),
    ("exponential", 3.0, 0, None),
    ("gamma", 2.0, 0, None),
    ("geometric", 0.35, 0, None),
    ("poisson", 5.0, 0, None),
    ("rayleigh", 2.0, 0, None),
    ("standard_gamma", 2.0, 0, None),
    ("weibull", 5.0, 0, None),
]


@pytest.fixture(params=list_of_two_arg_fn)
def two_arg_fn(request):
    return request.param


def get_two_arg_fn(op_name):
    return wrapper_function("a, b", f"np.random.{op_name}(a, b)", globals())


def test_two_arg_fn(filter_str, two_arg_fn, unary_size, capfd):
    if skip_test(filter_str):
        pytest.skip()

    op_name, first_arg, low, high = two_arg_fn

    if op_name == "gamma":
        pytest.skip("AttributeError: 'NoneType' object has no attribute 'ravel'")
    op = get_two_arg_fn(op_name)
    f = njit(op)
    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), dpnp_debug():
        actual = f(first_arg, unary_size)
        captured = capfd.readouterr()
        assert "dpnp implementation" in captured.out

        if low is not None and high is None:
            if np.isscalar(actual):
                assert actual >= low
            else:
                actual = actual.ravel()
                assert np.all(actual >= low)


list_of_three_arg_fn = [
    ("randint", 2, 23, 0, None),
    ("random_integers", 2, 23, 1, None),
    ("beta", 2.56, 0.8, 0, 1.0),
    ("binomial", 5, 0.0, 0, 1.0),
    ("gumbel", 0.5, 0.1, None, None),
    ("laplace", 0.0, 1.0, None, None),
    ("lognormal", 3.0, 1.0, None, None),
    ("multinomial", 100, np.array([1 / 7.0] * 5), 0, 100),
    ("multivariate_normal", (1, 2), [[1, 0], [0, 1]], None, None),
    ("negative_binomial", 1, 0.1, 0, None),
    ("normal", 0.0, 0.1, None, None),
    ("uniform", -1.0, 0.0, -1.0, 0.0),
]


@pytest.fixture(params=list_of_three_arg_fn)
def three_arg_fn(request):
    return request.param


def get_three_arg_fn(op_name):
    return wrapper_function("a, b, c", f"np.random.{op_name}(a, b, c)", globals())


def test_three_arg_fn(filter_str, three_arg_fn, three_arg_size, capfd):
    if skip_test(filter_str):
        pytest.skip()

    op_name, first_arg, second_arg, low, high = three_arg_fn

    if op_name == "multinomial":
        pytest.skip("DPNP RNG Error: dpnp_rng_multinomial_c() failed")
    elif op_name == "multivariate_normal":
        pytest.skip(
            "No implementation of function Function(<class "
            "'numba_dppy.dpnp_glue.stubs.dpnp.multivariate_normal'>) found for signature"
        )
    elif op_name == "negative_binomial":
        pytest.skip("DPNP RNG Error: dpnp_rng_negative_binomial_c() failed.")
    elif op_name == "gumbel":
        pytest.skip("DPNP error")

    op = get_three_arg_fn(op_name)
    f = njit(op)
    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), dpnp_debug():
        actual = f(first_arg, second_arg, three_arg_size)
        captured = capfd.readouterr()
        assert "dpnp implementation" in captured.out

        if low is not None and high is None:
            if second_arg:
                low = first_arg
                high = second_arg
                assert np.all(actual >= low)
                assert np.all(actual <= high)
            else:
                high = first_arg
                assert np.all(actual >= low)
                assert np.all(actual <= high)
        elif low is not None and high is not None:
            if np.isscalar(actual):
                assert actual >= low
                assert actual <= high
            else:
                actual = actual.ravel()
                assert np.all(actual >= low)
                assert np.all(actual <= high)


def test_rand(filter_str):
    if skip_test(filter_str):
        pytest.skip()

    @njit
    def f():
        c = np.random.rand(3, 2)
        return c

    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), dpnp_debug():
        actual = f()

        actual = actual.ravel()
        assert np.all(actual >= 0.0)
        assert np.all(actual < 1.0)


def test_hypergeometric(filter_str, three_arg_size):
    if skip_test(filter_str):
        pytest.skip()

    @njit
    def f(ngood, nbad, nsamp, size):
        res = np.random.hypergeometric(ngood, nbad, nsamp, size)
        return res

    ngood, nbad, nsamp = 100, 2, 10
    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), dpnp_debug():
        actual = f(ngood, nbad, nsamp, three_arg_size)

        if np.isscalar(actual):
            assert actual >= 0
            assert actual <= min(nsamp, ngood + nbad)
        else:
            actual = actual.ravel()
            assert np.all(actual >= 0)
            assert np.all(actual <= min(nsamp, ngood + nbad))



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import operator
from functools import reduce

import dpctl
import llvmlite.binding as ll
import llvmlite.llvmpy.core as lc
from llvmlite import ir
from llvmlite.llvmpy.core import Type
from numba.core import cgutils, types
from numba.core.imputils import Registry
from numba.core.itanium_mangler import mangle, mangle_c, mangle_type
from numba.core.typing.npydecl import parse_dtype

from numba_dppy import config, target
from numba_dppy.codegen import SPIR_DATA_LAYOUT
from numba_dppy.dppy_array_type import DPPYArray
from numba_dppy.ocl.atomics import atomic_helper
from numba_dppy.utils import address_space

from . import stubs

registry = Registry()
lower = registry.lower

_void_value = lc.Constant.null(lc.Type.pointer(lc.Type.int(8)))

# -----------------------------------------------------------------------------


def _declare_function(context, builder, name, sig, cargs, mangler=mangle_c):
    """Insert declaration for a opencl builtin function.
    Uses the Itanium mangler.

    Args
    ----
    context: target context

    builder: llvm builder

    name: str
        symbol name

    sig: signature
        function signature of the symbol being declared

    cargs: sequence of str
        C type names for the arguments

    mangler: a mangler function
        function to use to mangle the symbol

    """
    mod = builder.module
    if sig.return_type == types.void:
        llretty = lc.Type.void()
    else:
        llretty = context.get_value_type(sig.return_type)
    llargs = [context.get_value_type(t) for t in sig.args]
    fnty = Type.function(llretty, llargs)
    mangled = mangler(name, cargs)
    fn = cgutils.get_or_insert_function(mod, fnty, mangled)
    fn.calling_convention = target.CC_SPIR_FUNC
    return fn


@lower(stubs.get_global_id, types.uint32)
def get_global_id_impl(context, builder, sig, args):
    [dim] = args
    get_global_id = _declare_function(
        context, builder, "get_global_id", sig, ["unsigned int"]
    )
    res = builder.call(get_global_id, [dim])
    return context.cast(builder, res, types.uintp, types.intp)


@lower(stubs.get_local_id, types.uint32)
def get_local_id_impl(context, builder, sig, args):
    [dim] = args
    get_local_id = _declare_function(
        context, builder, "get_local_id", sig, ["unsigned int"]
    )
    res = builder.call(get_local_id, [dim])
    return context.cast(builder, res, types.uintp, types.intp)


@lower(stubs.get_group_id, types.uint32)
def get_group_id_impl(context, builder, sig, args):
    [dim] = args
    get_group_id = _declare_function(
        context, builder, "get_group_id", sig, ["unsigned int"]
    )
    res = builder.call(get_group_id, [dim])
    return context.cast(builder, res, types.uintp, types.intp)


@lower(stubs.get_num_groups, types.uint32)
def get_num_groups_impl(context, builder, sig, args):
    [dim] = args
    get_num_groups = _declare_function(
        context, builder, "get_num_groups", sig, ["unsigned int"]
    )
    res = builder.call(get_num_groups, [dim])
    return context.cast(builder, res, types.uintp, types.intp)


@lower(stubs.get_work_dim)
def get_work_dim_impl(context, builder, sig, args):
    get_work_dim = _declare_function(context, builder, "get_work_dim", sig, ["void"])
    res = builder.call(get_work_dim, [])
    return res


@lower(stubs.get_global_size, types.uint32)
def get_global_size_impl(context, builder, sig, args):
    [dim] = args
    get_global_size = _declare_function(
        context, builder, "get_global_size", sig, ["unsigned int"]
    )
    res = builder.call(get_global_size, [dim])
    return context.cast(builder, res, types.uintp, types.intp)


@lower(stubs.get_local_size, types.uint32)
def get_local_size_impl(context, builder, sig, args):
    [dim] = args
    get_local_size = _declare_function(
        context, builder, "get_local_size", sig, ["unsigned int"]
    )
    res = builder.call(get_local_size, [dim])
    return context.cast(builder, res, types.uintp, types.intp)


@lower(stubs.barrier, types.uint32)
def barrier_one_arg_impl(context, builder, sig, args):
    [flags] = args
    barrier = _declare_function(context, builder, "barrier", sig, ["unsigned int"])
    builder.call(barrier, [flags])
    return _void_value


@lower(stubs.barrier)
def barrier_no_arg_impl(context, builder, sig, args):
    assert not args
    sig = types.void(types.uint32)
    barrier = _declare_function(context, builder, "barrier", sig, ["unsigned int"])
    flags = context.get_constant(types.uint32, stubs.CLK_GLOBAL_MEM_FENCE)
    builder.call(barrier, [flags])
    return _void_value


@lower(stubs.mem_fence, types.uint32)
def mem_fence_impl(context, builder, sig, args):
    [flags] = args
    mem_fence = _declare_function(context, builder, "mem_fence", sig, ["unsigned int"])
    builder.call(mem_fence, [flags])
    return _void_value


@lower(stubs.sub_group_barrier)
def sub_group_barrier_impl(context, builder, sig, args):
    assert not args
    sig = types.void(types.uint32)
    barrier = _declare_function(context, builder, "barrier", sig, ["unsigned int"])
    flags = context.get_constant(types.uint32, stubs.CLK_LOCAL_MEM_FENCE)
    builder.call(barrier, [flags])
    return _void_value


def insert_and_call_atomic_fn(
    context, builder, sig, fn_type, dtype, ptr, val, addrspace
):
    ll_p = None
    name = ""
    if dtype.name == "float32":
        ll_val = ir.FloatType()
        ll_p = ll_val.as_pointer()
        if fn_type == "add":
            name = "numba_dppy_atomic_add_f32"
        elif fn_type == "sub":
            name = "numba_dppy_atomic_sub_f32"
        else:
            raise TypeError("Operation type is not supported %s" % (fn_type))
    elif dtype.name == "float64":
        if True:
            ll_val = ir.DoubleType()
            ll_p = ll_val.as_pointer()
            if fn_type == "add":
                name = "numba_dppy_atomic_add_f64"
            elif fn_type == "sub":
                name = "numba_dppy_atomic_sub_f64"
            else:
                raise TypeError("Operation type is not supported %s" % (fn_type))
    else:
        raise TypeError("Atomic operation is not supported for type %s" % (dtype.name))

    if addrspace == address_space.LOCAL:
        name = name + "_local"
    else:
        name = name + "_global"

    assert ll_p is not None
    assert name != ""
    ll_p.addrspace = address_space.GENERIC

    mod = builder.module
    if sig.return_type == types.void:
        llretty = lc.Type.void()
    else:
        llretty = context.get_value_type(sig.return_type)

    llargs = [ll_p, context.get_value_type(sig.args[2])]
    fnty = ir.FunctionType(llretty, llargs)

    fn = cgutils.get_or_insert_function(mod, fnty, name)
    fn.calling_convention = target.CC_SPIR_FUNC

    generic_ptr = context.addrspacecast(builder, ptr, address_space.GENERIC)

    return builder.call(fn, [generic_ptr, val])


def native_atomic_add(context, builder, sig, args):
    aryty, indty, valty = sig.args
    ary, inds, val = args
    dtype = aryty.dtype

    if indty == types.intp:
        indices = [inds]  # just a single integer
        indty = [indty]
    else:
        indices = cgutils.unpack_tuple(builder, inds, count=len(indty))
        indices = [
            context.cast(builder, i, t, types.intp) for t, i in zip(indty, indices)
        ]

    if dtype != valty:
        raise TypeError("expecting %s but got %s" % (dtype, valty))

    if aryty.ndim != len(indty):
        raise TypeError(
            "indexing %d-D array with %d-D index" % (aryty.ndim, len(indty))
        )

    lary = context.make_array(aryty)(context, builder, ary)
    ptr = cgutils.get_item_pointer(context, builder, aryty, lary, indices)

    if dtype == types.float32 or dtype == types.float64:
        context.extra_compile_options[target.LLVM_SPIRV_ARGS] = [
            "--spirv-ext=+SPV_EXT_shader_atomic_float_add"
        ]
        name = "__spirv_AtomicFAddEXT"
    elif dtype == types.int32 or dtype == types.int64:
        name = "__spirv_AtomicIAdd"
    else:
        raise TypeError("Unsupported type")

    assert name != ""

    ptr_type = context.get_value_type(dtype).as_pointer()
    ptr_type.addrspace = aryty.addrspace

    retty = context.get_value_type(sig.return_type)
    spirv_fn_arg_types = [
        ptr_type,
        ir.IntType(32),
        ir.IntType(32),
        context.get_value_type(sig.args[2]),
    ]

    from numba_dppy import extended_numba_itanium_mangler as ext_itanium_mangler

    numba_ptr_ty = types.CPointer(dtype, addrspace=ptr_type.addrspace)
    mangled_fn_name = ext_itanium_mangler.mangle(
        name,
        [numba_ptr_ty, "__spv.Scope.Flag", "__spv.MemorySemanticsMask.Flag", valty],
    )

    fnty = ir.FunctionType(retty, spirv_fn_arg_types)
    fn = cgutils.get_or_insert_function(builder.module, fnty, mangled_fn_name)
    fn.calling_convention = target.CC_SPIR_FUNC

    sycl_memory_order = atomic_helper.sycl_memory_order.relaxed
    sycl_memory_scope = atomic_helper.sycl_memory_scope.device
    spirv_scope = atomic_helper.get_scope(sycl_memory_scope)
    spirv_memory_semantics_mask = atomic_helper.get_memory_semantics_mask(
        sycl_memory_order
    )
    fn_args = [
        ptr,
        context.get_constant(types.int32, spirv_scope),
        context.get_constant(types.int32, spirv_memory_semantics_mask),
        val,
    ]

    return builder.call(fn, fn_args)


@lower(stubs.atomic.add, types.Array, types.intp, types.Any)
@lower(stubs.atomic.add, types.Array, types.UniTuple, types.Any)
@lower(stubs.atomic.add, types.Array, types.Tuple, types.Any)
def atomic_add_tuple(context, builder, sig, args):
    device_type = dpctl.get_current_queue().sycl_device.device_type
    dtype = sig.args[0].dtype

    if dtype == types.float32 or dtype == types.float64:
        if device_type == dpctl.device_type.gpu and config.NATIVE_FP_ATOMICS == 1:
            return native_atomic_add(context, builder, sig, args)
        else:
            # Currently, DPCPP only supports native floating point
            # atomics for GPUs.
            return atomic_add(context, builder, sig, args, "add")
    elif dtype == types.int32 or dtype == types.int64:
        return native_atomic_add(context, builder, sig, args)
    else:
        raise TypeError("Atomic operation on unsupported type %s" % dtype)


def atomic_sub_wrapper(context, builder, sig, args):
    # dpcpp yet does not support ``__spirv_AtomicFSubEXT``. To support atomic.sub we
    # reuse atomic.add and negate the value. For example, atomic.add(A, index, -val) is
    # equivalent to atomic.sub(A, index, val).
    val = args[2]
    new_val = cgutils.alloca_once(
        builder,
        context.get_value_type(sig.args[2]),
        size=context.get_constant(types.uintp, 1),
        name="new_val_0",
    )
    val_dtype = sig.args[2]
    if val_dtype == types.float32 or val_dtype == types.float64:
        builder.store(builder.fmul(val, context.get_constant(sig.args[2], -1)), new_val)
    elif val_dtype == types.int32 or val_dtype == types.int64:
        builder.store(builder.mul(val, context.get_constant(sig.args[2], -1)), new_val)
    else:
        raise TypeError("Unsupported type %s" % val_dtype)

    args[2] = builder.load(new_val)

    return native_atomic_add(context, builder, sig, args)


@lower(stubs.atomic.sub, types.Array, types.intp, types.Any)
@lower(stubs.atomic.sub, types.Array, types.UniTuple, types.Any)
@lower(stubs.atomic.sub, types.Array, types.Tuple, types.Any)
def atomic_sub_tuple(context, builder, sig, args):
    device_type = dpctl.get_current_queue().sycl_device.device_type
    dtype = sig.args[0].dtype

    if dtype == types.float32 or dtype == types.float64:
        if device_type == dpctl.device_type.gpu and config.NATIVE_FP_ATOMICS == 1:
            return atomic_sub_wrapper(context, builder, sig, args)
        else:
            # Currently, DPCPP only supports native floating point
            # atomics for GPUs.
            return atomic_add(context, builder, sig, args, "sub")
    elif dtype == types.int32 or dtype == types.int64:
        return atomic_sub_wrapper(context, builder, sig, args)
    else:
        raise TypeError("Atomic operation on unsupported type %s" % dtype)


def atomic_add(context, builder, sig, args, name):
    from .atomics import atomic_support_present

    if atomic_support_present():
        context.extra_compile_options[target.LINK_ATOMIC] = True
        aryty, indty, valty = sig.args
        ary, inds, val = args
        dtype = aryty.dtype

        if indty == types.intp:
            indices = [inds]  # just a single integer
            indty = [indty]
        else:
            indices = cgutils.unpack_tuple(builder, inds, count=len(indty))
            indices = [
                context.cast(builder, i, t, types.intp) for t, i in zip(indty, indices)
            ]

        if dtype != valty:
            raise TypeError("expecting %s but got %s" % (dtype, valty))

        if aryty.ndim != len(indty):
            raise TypeError(
                "indexing %d-D array with %d-D index" % (aryty.ndim, len(indty))
            )

        lary = context.make_array(aryty)(context, builder, ary)
        ptr = cgutils.get_item_pointer(context, builder, aryty, lary, indices)

        if isinstance(aryty, DPPYArray) and aryty.addrspace == address_space.LOCAL:
            return insert_and_call_atomic_fn(
                context,
                builder,
                sig,
                name,
                dtype,
                ptr,
                val,
                address_space.LOCAL,
            )
        else:
            return insert_and_call_atomic_fn(
                context,
                builder,
                sig,
                name,
                dtype,
                ptr,
                val,
                address_space.GLOBAL,
            )
    else:
        raise ImportError("Atomic support is not present, can not perform atomic_add")


@lower(stubs.local.array, types.IntegerLiteral, types.Any)
def dppy_local_array_integer(context, builder, sig, args):
    length = sig.args[0].literal_value
    dtype = parse_dtype(sig.args[1])
    return _generic_array(
        context,
        builder,
        shape=(length,),
        dtype=dtype,
        symbol_name="_dppy_lmem",
        addrspace=address_space.LOCAL,
    )


@lower(stubs.local.array, types.Tuple, types.Any)
@lower(stubs.local.array, types.UniTuple, types.Any)
def dppy_local_array_tuple(context, builder, sig, args):
    shape = [s.literal_value for s in sig.args[0]]
    dtype = parse_dtype(sig.args[1])
    return _generic_array(
        context,
        builder,
        shape=shape,
        dtype=dtype,
        symbol_name="_dppy_lmem",
        addrspace=address_space.LOCAL,
    )


def _generic_array(context, builder, shape, dtype, symbol_name, addrspace):
    """
    This function allows us to create generic arrays in different
    address spaces.
    """
    elemcount = reduce(operator.mul, shape)
    lldtype = context.get_data_type(dtype)
    laryty = Type.array(lldtype, elemcount)

    if addrspace == address_space.LOCAL:
        lmod = builder.module

        # Create global variable in the requested address-space
        gvmem = lmod.add_global_variable(laryty, symbol_name, addrspace)

        if elemcount <= 0:
            raise ValueError("array length <= 0")
        else:
            gvmem.linkage = lc.LINKAGE_INTERNAL

        if dtype not in types.number_domain:
            raise TypeError("unsupported type: %s" % dtype)

    else:
        raise NotImplementedError("addrspace {addrspace}".format(**locals()))

    # We need to add the addrspace to _make_array() function call as we want
    # the variable containing the reference of the memory to retain the
    # original address space of that memory. Before, we were casting the
    # memories allocated in local address space to global address space. This
    # approach does not let us identify the original address space of a memory
    # down the line.
    return _make_array(context, builder, gvmem, dtype, shape, addrspace=addrspace)


def _make_array(
    context,
    builder,
    dataptr,
    dtype,
    shape,
    layout="C",
    addrspace=address_space.GENERIC,
):
    ndim = len(shape)
    # Create array object
    aryty = DPPYArray(dtype=dtype, ndim=ndim, layout="C", addrspace=addrspace)
    ary = context.make_array(aryty)(context, builder)

    targetdata = _get_target_data(context)
    lldtype = context.get_data_type(dtype)
    itemsize = lldtype.get_abi_size(targetdata)
    # Compute strides
    rstrides = [itemsize]
    for i, lastsize in enumerate(reversed(shape[1:])):
        rstrides.append(lastsize * rstrides[-1])
    strides = [s for s in reversed(rstrides)]

    kshape = [context.get_constant(types.intp, s) for s in shape]
    kstrides = [context.get_constant(types.intp, s) for s in strides]

    context.populate_array(
        ary,
        data=builder.bitcast(dataptr, ary.data.type),
        shape=cgutils.pack_array(builder, kshape),
        strides=cgutils.pack_array(builder, kstrides),
        itemsize=context.get_constant(types.intp, itemsize),
        meminfo=None,
    )

    return ary._getvalue()


def _get_target_data(context):
    return ll.create_target_data(SPIR_DATA_LAYOUT[context.address_size])



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import dpctl
import numpy as np
import pytest
from numba import njit
from numba.tests.support import captured_stdout

import numba_dppy as dppy
from numba_dppy import config

from . import _helper
from ._helper import assert_auto_offloading

skip_no_gpu = pytest.mark.skipif(
    not _helper.has_gpu_queues("opencl"),
    reason="No opencl GPU platforms available",
)
skip_no_cpu = pytest.mark.skipif(
    not _helper.has_cpu_queues("opencl"),
    reason="No opencl CPU platforms available",
)
skip_no_level_zero = pytest.mark.skipif(
    not _helper.has_gpu_queues("level_zero"),
    reason="No level-zero GPU platforms available",
)

filter_strings = [
    pytest.param("level_zero:gpu:0", marks=skip_no_level_zero),
    pytest.param("opencl:gpu:0", marks=skip_no_gpu),
    pytest.param("opencl:cpu:0", marks=skip_no_cpu),
]


def scenario(filter_str, context):
    @njit
    def func(a, b):
        return a + b

    a = np.ones((64), dtype=np.int32)
    b = np.ones((64), dtype=np.int32)

    device = dpctl.SyclDevice(filter_str)
    with context(device):
        func(a, b)


@pytest.mark.parametrize("filter_str", filter_strings)
@pytest.mark.parametrize(
    "context",
    [
        dppy.offload_to_sycl_device,
        dpctl.device_context,
    ],
)
def test_dpctl_device_context_affects_numba_pipeline(filter_str, context):
    with assert_auto_offloading():
        scenario(filter_str, context)


class TestWithDPPYContext(unittest.TestCase):
    @unittest.skipIf(not _helper.has_gpu_queues(), "No GPU platforms available")
    def test_with_dppy_context_gpu(self):
        @njit
        def nested_func(a, b):
            np.sin(a, b)

        @njit
        def func(b):
            a = np.ones((64), dtype=np.float64)
            nested_func(a, b)

        config.DEBUG = 1
        expected = np.ones((64), dtype=np.float64)
        got_gpu = np.ones((64), dtype=np.float64)

        with captured_stdout() as got_gpu_message:
            device = dpctl.SyclDevice("opencl:gpu")
            with dpctl.device_context(device):
                func(got_gpu)

        config.DEBUG = 0
        func(expected)

        np.testing.assert_array_equal(expected, got_gpu)
        self.assertTrue("Parfor offloaded to opencl:gpu" in got_gpu_message.getvalue())

    @unittest.skipIf(not _helper.has_cpu_queues(), "No CPU platforms available")
    def test_with_dppy_context_cpu(self):
        @njit
        def nested_func(a, b):
            np.sin(a, b)

        @njit
        def func(b):
            a = np.ones((64), dtype=np.float64)
            nested_func(a, b)

        config.DEBUG = 1
        expected = np.ones((64), dtype=np.float64)
        got_cpu = np.ones((64), dtype=np.float64)

        with captured_stdout() as got_cpu_message:
            device = dpctl.SyclDevice("opencl:cpu")
            with dpctl.device_context(device):
                func(got_cpu)

        config.DEBUG = 0
        func(expected)

        np.testing.assert_array_equal(expected, got_cpu)
        self.assertTrue("Parfor offloaded to opencl:cpu" in got_cpu_message.getvalue())


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numba import types
from numba.core.typing.npydecl import parse_dtype, parse_shape
from numba.core.typing.templates import (
    AbstractTemplate,
    AttributeTemplate,
    CallableTemplate,
    ConcreteTemplate,
    Registry,
    signature,
)

import numba_dppy as dppy
from numba_dppy.dppy_array_type import DPPYArray
from numba_dppy.utils import address_space

registry = Registry()
intrinsic = registry.register
intrinsic_attr = registry.register_attr
intrinsic_global = registry.register_global

# register_number_classes(intrinsic_global)


@intrinsic
class Ocl_get_global_id(ConcreteTemplate):
    key = dppy.get_global_id
    cases = [signature(types.intp, types.uint32)]


@intrinsic
class Ocl_get_local_id(ConcreteTemplate):
    key = dppy.get_local_id
    cases = [signature(types.intp, types.uint32)]


@intrinsic
class Ocl_get_group_id(ConcreteTemplate):
    key = dppy.get_group_id
    cases = [signature(types.intp, types.uint32)]


@intrinsic
class Ocl_get_num_groups(ConcreteTemplate):
    key = dppy.get_num_groups
    cases = [signature(types.intp, types.uint32)]


@intrinsic
class Ocl_get_work_dim(ConcreteTemplate):
    key = dppy.get_work_dim
    cases = [signature(types.uint32)]


@intrinsic
class Ocl_get_global_size(ConcreteTemplate):
    key = dppy.get_global_size
    cases = [signature(types.intp, types.uint32)]


@intrinsic
class Ocl_get_local_size(ConcreteTemplate):
    key = dppy.get_local_size
    cases = [signature(types.intp, types.uint32)]


@intrinsic
class Ocl_barrier(ConcreteTemplate):
    key = dppy.barrier
    cases = [signature(types.void, types.uint32), signature(types.void)]


@intrinsic
class Ocl_mem_fence(ConcreteTemplate):
    key = dppy.mem_fence
    cases = [signature(types.void, types.uint32)]


@intrinsic
class Ocl_sub_group_barrier(ConcreteTemplate):
    key = dppy.sub_group_barrier

    cases = [signature(types.void)]


# dppy.atomic submodule -------------------------------------------------------


@intrinsic
class Ocl_atomic_add(AbstractTemplate):
    key = dppy.atomic.add

    def generic(self, args, kws):
        assert not kws
        ary, idx, val = args

        if ary.ndim == 1:
            return signature(ary.dtype, ary, types.intp, ary.dtype)
        elif ary.ndim > 1:
            return signature(ary.dtype, ary, idx, ary.dtype)


@intrinsic
class Ocl_atomic_sub(AbstractTemplate):
    key = dppy.atomic.sub

    def generic(self, args, kws):
        assert not kws
        ary, idx, val = args

        if ary.ndim == 1:
            return signature(ary.dtype, ary, types.intp, ary.dtype)
        elif ary.ndim > 1:
            return signature(ary.dtype, ary, idx, ary.dtype)


@intrinsic_attr
class OclAtomicTemplate(AttributeTemplate):
    key = types.Module(dppy.atomic)

    def resolve_add(self, mod):
        return types.Function(Ocl_atomic_add)

    def resolve_sub(self, mod):
        return types.Function(Ocl_atomic_sub)


intrinsic_global(dppy.atomic.add, types.Function(Ocl_atomic_add))
intrinsic_global(dppy.atomic.sub, types.Function(Ocl_atomic_sub))

# dppy.local submodule -------------------------------------------------------


@intrinsic
class OCL_local_array(CallableTemplate):
    key = dppy.local.array

    def generic(self):
        def typer(shape, dtype):

            # Only integer literals and tuples of integer literals are valid
            # shapes
            if isinstance(shape, types.Integer):
                if not isinstance(shape, types.IntegerLiteral):
                    return None
            elif isinstance(shape, (types.Tuple, types.UniTuple)):
                if any([not isinstance(s, types.IntegerLiteral) for s in shape]):
                    return None
            else:
                return None

            ndim = parse_shape(shape)
            nb_dtype = parse_dtype(dtype)
            if nb_dtype is not None and ndim is not None:
                return DPPYArray(
                    dtype=nb_dtype,
                    ndim=ndim,
                    layout="C",
                    addrspace=address_space.LOCAL,
                )

        return typer


@intrinsic_attr
class OclLocalTemplate(AttributeTemplate):
    key = types.Module(dppy.local)

    def resolve_array(self, mod):
        return types.Function(OCL_local_array)


# OpenCL module --------------------------------------------------------------


@intrinsic_attr
class OclModuleTemplate(AttributeTemplate):
    key = types.Module(dppy)

    def resolve_get_global_id(self, mod):
        return types.Function(Ocl_get_global_id)

    def resolve_get_local_id(self, mod):
        return types.Function(Ocl_get_local_id)

    def resolve_get_global_size(self, mod):
        return types.Function(Ocl_get_global_size)

    def resolve_get_local_size(self, mod):
        return types.Function(Ocl_get_local_size)

    def resolve_get_num_groups(self, mod):
        return types.Function(Ocl_get_num_groups)

    def resolve_get_work_dim(self, mod):
        return types.Function(Ocl_get_work_dim)

    def resolve_get_group_id(self, mod):
        return types.Function(Ocl_get_group_id)

    def resolve_barrier(self, mod):
        return types.Function(Ocl_barrier)

    def resolve_mem_fence(self, mod):
        return types.Function(Ocl_mem_fence)

    def resolve_sub_group_barrier(self, mod):
        return types.Function(Ocl_sub_group_barrier)

    def resolve_atomic(self, mod):
        return types.Module(dppy.atomic)

    def resolve_local(self, mod):
        return types.Module(dppy.local)


# intrinsic

# intrinsic_global(dppy, types.Module(dppy))



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numba.core import types
from numba.core.datamodel.models import StructModel
from numba.core.types.npytypes import Array


class DPPYArray(Array):
    """
    Type class for DPPY arrays.
    """

    def __init__(
        self,
        dtype,
        ndim,
        layout,
        readonly=False,
        name=None,
        aligned=True,
        addrspace=None,
    ):
        self.addrspace = addrspace
        super(DPPYArray, self).__init__(
            dtype,
            ndim,
            layout,
            readonly=readonly,
            name=name,
            aligned=aligned,
        )

    def copy(self, dtype=None, ndim=None, layout=None, readonly=None, addrspace=None):
        if dtype is None:
            dtype = self.dtype
        if ndim is None:
            ndim = self.ndim
        if layout is None:
            layout = self.layout
        if readonly is None:
            readonly = not self.mutable
        if addrspace is None:
            addrspace = self.addrspace
        return DPPYArray(
            dtype=dtype,
            ndim=ndim,
            layout=layout,
            readonly=readonly,
            aligned=self.aligned,
            addrspace=addrspace,
        )

    @property
    def key(self):
        return (
            self.dtype,
            self.ndim,
            self.layout,
            self.mutable,
            self.aligned,
            self.addrspace,
        )

    @property
    def box_type(self):
        return np.ndarray

    def is_precise(self):
        return self.dtype.is_precise()


class DPPYArrayModel(StructModel):
    def __init__(self, dmm, fe_type):
        ndim = fe_type.ndim
        members = [
            ("meminfo", types.CPointer(fe_type.dtype, addrspace=fe_type.addrspace)),
            ("parent", types.CPointer(fe_type.dtype, addrspace=fe_type.addrspace)),
            ("nitems", types.intp),
            ("itemsize", types.intp),
            ("data", types.CPointer(fe_type.dtype, addrspace=fe_type.addrspace)),
            ("shape", types.UniTuple(types.intp, ndim)),
            ("strides", types.UniTuple(types.intp, ndim)),
        ]
        super(DPPYArrayModel, self).__init__(dmm, fe_type, members)



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import warnings

import dpctl
import numba
import numpy as np

from . import _helper


@unittest.skipUnless(_helper.has_gpu_queues(), "test only on GPU system")
class TestDPPYFallback(unittest.TestCase):
    def test_dppy_fallback_inner_call(self):
        @numba.jit
        def fill_value(i):
            return i

        def inner_call_fallback():
            x = 10
            a = np.empty(shape=x, dtype=np.float32)

            for i in numba.prange(x):
                a[i] = fill_value(i)

            return a

        device = dpctl.SyclDevice("opencl:gpu")
        with warnings.catch_warnings(record=True) as w, dpctl.device_context(device):
            dppy = numba.njit(inner_call_fallback)
            dppy_result = dppy()

        ref_result = inner_call_fallback()

        np.testing.assert_array_equal(dppy_result, ref_result)
        self.assertIn("Failed to offload parfor ", str(w[-1].message))

    def test_dppy_fallback_reductions(self):
        def reduction(a):
            b = 1
            for i in numba.prange(len(a)):
                b += a[i]
            return b

        a = np.ones(10)
        device = dpctl.SyclDevice("opencl:gpu")
        with warnings.catch_warnings(record=True) as w, dpctl.device_context(device):
            dppy = numba.njit(reduction)
            dppy_result = dppy(a)

        ref_result = reduction(a)

        np.testing.assert_array_equal(dppy_result, ref_result)
        self.assertIn("Failed to offload parfor ", str(w[-1].message))


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numba.core import itanium_mangler, types


def mangle_type_or_value(typ):
    """
    Mangle type parameter and arbitrary value.

    This function extends Numba's `magle_type_or_value()` to
    support numba.types.CPointer type, e.g. an ``int *`` argument will be
    mangled to "Pi".
    Mangling of extended qualifiers is supported only
    for address space qualifiers. In which case, the mangling
    follows the rule defined in Section 5.1.5.1 of the ``Itanium ABI
    <https://itanium-cxx-abi.github.io/cxx-abi/abi.html#mangle.qualified-type>``_.
    For example, an ``int global *`` argument will be mangeled to "PU3AS1i".

    Args:
        typ (numba.types, int, str) : Type to mangle

    Returns:
        str: The mangled name of the type

    """
    if isinstance(typ, types.CPointer):
        rc = "P"
        if typ.addrspace is not None:
            rc += "U" + itanium_mangler.mangle_identifier("AS" + str(typ.addrspace))
        rc += itanium_mangler.mangle_type_or_value(typ.dtype)
        return rc
    else:
        return itanium_mangler.mangle_type_or_value(typ)


mangle_type = mangle_type_or_value


def mangle_args(argtys):
    """
    Mangle sequence of Numba type objects and arbitrary values.
    """
    return "".join([mangle_type_or_value(t) for t in argtys])


def mangle(ident, argtys):
    """
    Mangle identifier with Numba type objects and arbitrary values.
    """
    return (
        itanium_mangler.PREFIX
        + itanium_mangler.mangle_identifier(ident)
        + mangle_args(argtys)
    )



#! /usr/bin/env python
# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import numba
import numpy as np
from numba import typeof
from numba.core import compiler, cpu, typing
from numba.core.typed_passes import AnnotateTypes, NopythonTypeInference

import numba_dppy
from numba_dppy.rename_numpy_functions_pass import (
    DPPYRewriteNdarrayFunctions,
    DPPYRewriteOverloadedNumPyFunctions,
)
from numba_dppy.tests._helper import ensure_dpnp


class MyPipeline(object):
    def __init__(self, test_ir, args):
        self.state = compiler.StateDict()
        self.state.typingctx = typing.Context()
        self.state.targetctx = cpu.CPUContext(self.state.typingctx)
        self.state.func_ir = test_ir
        self.state.func_id = test_ir.func_id
        self.state.args = args
        self.state.return_type = None
        self.state.locals = dict()
        self.state.status = None
        self.state.lifted = dict()
        self.state.lifted_from = None
        self.state.typingctx.refresh()
        self.state.targetctx.refresh()


def check_equivalent(expected_ir, got_ir):
    expected_block_body = expected_ir.blocks[0].body
    got_block_body = got_ir.blocks[0].body

    if len(expected_block_body) != len(got_block_body):
        return False

    for i in range(len(expected_block_body)):
        expected_stmt = expected_block_body[i]
        got_stmt = got_block_body[i]
        if type(expected_stmt) != type(got_stmt):
            return False
        else:
            if isinstance(expected_stmt, numba.core.ir.Assign):
                if isinstance(expected_stmt.value, numba.core.ir.Global):
                    if (
                        expected_stmt.value.name != got_stmt.value.name
                        and expected_stmt.value.name != "numba_dppy"
                    ):
                        return False
                elif isinstance(expected_stmt.value, numba.core.ir.Expr):
                    # should get "dpnp" and "sum" as attr
                    if expected_stmt.value.op == "getattr":
                        if expected_stmt.value.attr != got_stmt.value.attr:
                            return False
    return True


class TestRenameNumpyFunctionsPass(unittest.TestCase):
    def test_rename_numpy(self):
        def expected(a):
            return numba_dppy.dpnp.sum(a)

        def got(a):
            return np.sum(a)

        expected_ir = compiler.run_frontend(expected)
        got_ir = compiler.run_frontend(got)

        pipeline = MyPipeline(got_ir, None)

        rewrite_numpy_functions_pass = DPPYRewriteOverloadedNumPyFunctions()
        rewrite_numpy_functions_pass.run_pass(pipeline.state)

        self.assertTrue(check_equivalent(expected_ir, pipeline.state.func_ir))


@unittest.skipUnless(ensure_dpnp(), "test only when dpnp is available")
class TestRenameNdarrayFunctionsPass(unittest.TestCase):
    def test_rename_ndarray(self):
        def expected(a):
            return numba_dppy.dpnp.sum(a)

        def got(a):
            return a.sum()

        expected_ir = compiler.run_frontend(expected)
        got_ir = compiler.run_frontend(got)

        a = np.arange(10)
        args = [a]
        argtypes = [typeof(x) for x in args]

        pipeline = MyPipeline(got_ir, argtypes)

        tyinfer_pass = NopythonTypeInference()
        tyinfer_pass.run_pass(pipeline.state)

        annotate_ty_pass = AnnotateTypes()
        annotate_ty_pass.run_pass(pipeline.state)

        rewrite_ndarray_functions_pass = DPPYRewriteNdarrayFunctions()
        rewrite_ndarray_functions_pass.run_pass(pipeline.state)

        self.assertTrue(check_equivalent(expected_ir, pipeline.state.func_ir))


if __name__ == "__main__":
    unittest.main()



################################################################################
#                                 Numba-DPPY
#
# Copyright 2020-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import string


def wrapper_function(args, code, globals, function_name="func"):
    function_text = f"""\
def {function_name}({args}):
    return {code}
"""
    return compile_function(function_text, function_name, globals)


def args_string(args_count):
    return ", ".join(list(string.ascii_lowercase[:args_count]))


def compile_function(function_text, function_name, globals):
    locals = {}
    exec(function_text, globals, locals)
    return locals[function_name]



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides a set of wrapper functions to insert dpctl C API function
declarations into an LLVM module.
"""

import llvmlite.llvmpy.core as lc
from numba.core import cgutils, types

import numba_dppy.utils as utils


class DpctlCAPIFnBuilder:
    """
    Defines a set of static functions to add declarations for dpctl C API
    library function into an LLVM module.
    """

    @staticmethod
    def _build_dpctl_function(builder, return_ty, arg_list, func_name):
        """
        _build_dpctl_function(builder, return_ty, arg_list, func_name)
        Inserts an LLVM function of the specified signature and name into an
        LLVM module.

        Args:
            return_ty: An LLVM Value corresponding to the return type of the
                       function.
            arg_list: A list of LLVM Value objects corresponsing to the
                      type of the arguments for the function.
            func_name: The name of the function passed as a string.

        Return: A Python object wrapping an LLVM Function.

        """
        func_ty = lc.Type.function(return_ty, arg_list)
        fn = cgutils.get_or_insert_function(builder.module, func_ty, func_name)
        return fn

    @staticmethod
    def get_dpctl_queuemgr_get_current_queue(builder, context):
        """
        get_dpctl_queuemgr_get_current_queue(builder, context)
        Inserts an LLVM Function for ``DPCTLQueueMgr_GetCurrentQueue``.

        The ``DPCTLQueueMgr_GetCurrentQueue`` function returns the current
        top-of-the-stack SYCL queue that is stored in dpclt's queue manager.

        Args:
            builder: The LLVM IR builder to be used for code generation.
            context: The LLVM IR builder context.

        Return: A Python object wrapping an LLVM Function for
                ``DPCTLQueueMgr_GetCurrentQueue``.
        """
        return DpctlCAPIFnBuilder._build_dpctl_function(
            builder,
            return_ty=utils.get_llvm_type(context=context, type=types.voidptr),
            arg_list=[],
            func_name="DPCTLQueueMgr_GetCurrentQueue",
        )

    @staticmethod
    def get_dpctl_queue_delete(builder, context):
        """
        get_dpctl_queue_delete(builder, context)
        Inserts an LLVM Function for the ``DPCTLQueue_Delete``.

        The ``DPCTLQueue_Delete`` deletes a DPCTLSyclQueueRef opaque pointer.

        Args:
            builder: The LLVM IR builder to be used for code generation.
            context: The LLVM IR builder context.

        Return: A Python object wrapping an LLVM Function for
                ``DPCTLQueue_Delete``.
        """
        return DpctlCAPIFnBuilder._build_dpctl_function(
            builder,
            return_ty=utils.LLVMTypes.void_t,
            arg_list=[utils.get_llvm_type(context=context, type=types.voidptr)],
            func_name="DPCTLQueue_Delete",
        )

    @staticmethod
    def get_dpctl_queue_memcpy(builder, context):
        """
        get_dpctl_queue_memcpy(builder, context)
        Inserts an LLVM Function for the ``DPCTLQueue_Memcpy``.

        The ``DPCTLQueue_Memcpy`` function is a wrapper over ``sycl::queue``
        class' ``event memcpy(void* dest, const void* src, size_t numBytes)``
        function. Currently, the DPCTLQueue_Memcpy does not return an event
        reference, but in future will return an opaque pointer to a
        ``sycl::event``. All the opaque pointers arguments to the
        ``DPCTLQueue_Memcpy`` are passed as void pointers.

        Args:
            builder: The LLVM IR builder to be used for code generation.
            context: The LLVM IR builder context.

        Return: A Python object wrapping an LLVM Function for
                ``DPCTLQueue_Memcpy``.
        """
        void_ptr_t = utils.get_llvm_type(context=context, type=types.voidptr)
        return DpctlCAPIFnBuilder._build_dpctl_function(
            builder,
            return_ty=void_ptr_t,
            # return_ty=utils.LLVMTypes.void_t,
            arg_list=[
                void_ptr_t,
                void_ptr_t,
                void_ptr_t,
                utils.get_llvm_type(context=context, type=types.intp),
            ],
            func_name="DPCTLQueue_Memcpy",
        )

    @staticmethod
    def get_dpctl_queue_wait(builder, context):
        """
        get_dpctl_queue_wait(builder, context)
        Inserts an LLVM Function for ``DPCTLQueue_Wait``.

        The ``DPCTLQueue_Wait`` is a wrapper over ``sycl::queue`` class'
        ``wait()`` function.

        Args:
            builder: The LLVM IR builder to be used for code generation.
            context: The LLVM IR builder context.

        Returns: A Python object wrapping an LLVM Function for
                ``DPCTLQueue_Wait``.
        """
        void_ptr_t = utils.get_llvm_type(context=context, type=types.voidptr)
        return DpctlCAPIFnBuilder._build_dpctl_function(
            builder,
            return_ty=utils.LLVMTypes.void_t,
            arg_list=[void_ptr_t],
            func_name="DPCTLQueue_Wait",
        )

    @staticmethod
    def get_dpctl_queue_submit_range(builder, context):
        """
        get_dpctl_queue_submit_range(builder, context)
        Inserts an LLVM Function for ``DPCTLQueue_SubmitRange``.

        The ``DPCTLQueue_SubmitRange`` function is a wrapper over the
        ``sycl::queue`` class' ``event parallel_for(range<dimensions>
        numWorkItems, Rest&&... rest)`` function. All the opaque pointers
        arguments to the ``DPCTLQueue_SubmitRange`` function are passed as void
        pointers.

        Note: The ``DPCTLQueue_SubmitRange`` calls returns an opaque pointer to
        a ``sycl::event`` that needs to be destroyed properly.

        Args:
            builder: The LLVM IR builder to be used for code generation.
            context: The LLVM IR builder context.

        Return: A Python object wrapping an LLVM Function for
                ``DPCTLQueue_SubmitRange``.
        """
        intp_t = utils.get_llvm_type(context=context, type=types.intp)
        void_ptr_t = utils.get_llvm_type(context=context, type=types.voidptr)
        void_ptr_ptr_t = utils.get_llvm_ptr_type(void_ptr_t)
        intp_ptr_t = utils.get_llvm_ptr_type(intp_t)

        return DpctlCAPIFnBuilder._build_dpctl_function(
            builder,
            return_ty=void_ptr_t,
            arg_list=[
                void_ptr_t,
                void_ptr_t,
                void_ptr_ptr_t,
                utils.LLVMTypes.int32_ptr_t,
                utils.get_llvm_type(context=context, type=types.intp),
                intp_ptr_t,
                utils.get_llvm_type(context=context, type=types.intp),
                void_ptr_t,
                utils.get_llvm_type(context=context, type=types.intp),
            ],
            func_name="DPCTLQueue_SubmitRange",
        )

    @staticmethod
    def get_dpctl_malloc_shared(builder, context):
        """
        get_dpctl_malloc_shared(builder, context)
        Inserts an LLVM Function for ``DPCTLmalloc_shared``.

        ``DPCTLmalloc_shared`` is a wrapper over the
        ``sycl::malloc_shared`` function to allocate USM shared memory.

        Args:
            builder: The LLVM IR builder to be used for code generation.
            context: The LLVM IR builder context.

        Return: A Python object wrapping an LLVM Function for
                ``DPCTLmalloc_shared``.

        """
        void_ptr_t = utils.get_llvm_type(context=context, type=types.voidptr)
        int_ptr_t = utils.get_llvm_type(context=context, type=types.intp)
        return DpctlCAPIFnBuilder._build_dpctl_function(
            builder,
            return_ty=void_ptr_t,
            arg_list=[int_ptr_t, void_ptr_t],
            func_name="DPCTLmalloc_shared",
        )

    @staticmethod
    def get_dpctl_free_with_queue(builder, context):
        """
        get_dpctl_free_with_queue(builder, context)
        Inserts an LLVM Function for ``DPCTLfree_with_queue``.

        The ``DPCTLfree_with_queue`` function is a wrapper over
        ``sycl::free(void*, queue)`` function. All the opaque pointers
        arguments to the ``DPCTLfree_with_queue`` are passed as void pointers.

        Args:
            builder: The LLVM IR builder to be used for code generation.
            context: The LLVM IR builder context.

        Return: A Python object wrapping an LLVM Function for
                ``DPCTLfree_with_queue``.
        """
        void_ptr_t = utils.get_llvm_type(context=context, type=types.voidptr)
        return DpctlCAPIFnBuilder._build_dpctl_function(
            builder,
            return_ty=utils.LLVMTypes.void_t,
            arg_list=[void_ptr_t, void_ptr_t],
            func_name="DPCTLfree_with_queue",
        )

    @staticmethod
    def get_dpctl_event_wait(builder, context):
        """
        get_dpctl_event_wait(builder, context)
        Inserts an LLVM Function for ``DPCTLEvent_Wait``.

        The ``DPCTLEvent_Wait`` function is a wrapper over the
        ``sycl::event`` class' ``wait()`` function.

        Args:
            builder: The LLVM IR builder to be used for code generation.
            context: The LLVM IR builder context.

        Return: A Python object wrapping an LLVM Function for
                ``DPCTLEvent_Wait``.
        """
        void_ptr_t = utils.get_llvm_type(context=context, type=types.voidptr)
        return DpctlCAPIFnBuilder._build_dpctl_function(
            builder,
            return_ty=utils.LLVMTypes.void_t,
            arg_list=[void_ptr_t],
            func_name="DPCTLEvent_Wait",
        )

    @staticmethod
    def get_dpctl_event_delete(builder, context):
        """
        get_dpctl_event_delete(builder, context)
        Inserts an LLVM Function for ``DPCTLEvent_Delete``.

        The ``DPCTLEvent_Delete`` function deletes a ``DPCTLSyclEventRef``
        opaque pointer.

        Args:
            builder: The LLVM IR builder to be used for code generation.
            context: The LLVM IR builder context.

        Returns: A Python object wrapping an LLVM Function for
                 ``DPCTLEvent_Delete``.
        """
        void_ptr_t = utils.get_llvm_type(context=context, type=types.voidptr)
        return DpctlCAPIFnBuilder._build_dpctl_function(
            builder,
            return_ty=utils.LLVMTypes.void_t,
            arg_list=[void_ptr_t],
            func_name="DPCTLEvent_Delete",
        )



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dpctl
import dpctl.tensor.numpy_usm_shared as usmarray
import numpy as np
import pytest
from numba import prange
from numba.core import compiler, cpu
from numba.core.registry import cpu_target

import numba_dppy as dppy
from numba_dppy.compiler import DPPYCompiler


def fn(a):
    for i in prange(a.size):
        a[i] += 1
    return a


def test_no_copy_usm_shared(capfd):
    a = usmarray.ones(10, dtype=np.int64)
    b = np.ones(10, dtype=np.int64)
    # f = njit(fn)

    flags = compiler.Flags()
    flags.no_compile = True
    flags.no_cpython_wrapper = True
    flags.nrt = False
    flags.auto_parallel = cpu.ParallelOptions(True)

    typingctx = cpu_target.typing_context
    targetctx = cpu_target.target_context
    args = typingctx.resolve_argument_type(a)

    try:
        device = dpctl.SyclDevice("opencl:gpu:0")
    except ValueError:
        pytest.skip("Device not found")

    with dpctl.device_context(device):
        cres = compiler.compile_extra(
            typingctx=typingctx,
            targetctx=targetctx,
            func=fn,
            args=tuple([args]),
            return_type=args,
            flags=flags,
            locals={},
            pipeline_class=DPPYCompiler,
        )

        assert "DPCTLQueue_Memcpy" not in cres.library.get_llvm_str()

        args = typingctx.resolve_argument_type(b)
        cres = compiler.compile_extra(
            typingctx=typingctx,
            targetctx=targetctx,
            func=fn,
            args=tuple([args]),
            return_type=args,
            flags=flags,
            locals={},
            pipeline_class=DPPYCompiler,
        )

        assert "DPCTLQueue_Memcpy" in cres.library.get_llvm_str()



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numba.core import compiler, dispatcher
from numba.core.registry import cpu_target
from numba.core.target_extension import dispatcher_registry, target_registry

import numba_dppy.config as dppy_config
from numba_dppy.target import DPPY_TARGET_NAME


class DppyOffloadDispatcher(dispatcher.Dispatcher):
    targetdescr = cpu_target

    def __init__(
        self,
        py_func,
        locals={},
        targetoptions={},
        impl_kind="direct",
        pipeline_class=compiler.Compiler,
    ):
        if dppy_config.dppy_present:
            from numba_dppy.compiler import DPPYCompiler

            targetoptions["parallel"] = True
            dispatcher.Dispatcher.__init__(
                self,
                py_func,
                locals=locals,
                targetoptions=targetoptions,
                impl_kind=impl_kind,
                pipeline_class=DPPYCompiler,
            )
        else:
            print(
                "---------------------------------------------------------------------"
            )
            print(
                "WARNING : DPPY pipeline ignored. Ensure OpenCL drivers are installed."
            )
            print(
                "---------------------------------------------------------------------"
            )
            dispatcher.Dispatcher.__init__(
                self,
                py_func,
                locals=locals,
                targetoptions=targetoptions,
                impl_kind=impl_kind,
                pipeline_class=pipeline_class,
            )


dispatcher_registry[target_registry[DPPY_TARGET_NAME]] = DppyOffloadDispatcher
